import re
from itertools import product

from app.readers import pycolator as reader
from app.readers import fasta
from app.readers import xmlformatting as formatting


def get_either_seq(seqtype, element, ns):
    get_seq_map = {'psm': reader.get_psm_seq,
                   'pep': reader.get_peptide_seq,
                   }
    return get_seq_map[seqtype](element, ns)


def filter_peptide_length(features, elementtype, ns, minlen=0, maxlen=None):
    minlen = int(minlen)
    if maxlen is None:
        maxlen = float('inf')
    else:
        maxlen = int(maxlen)
    for feat in features:
        seq = get_either_seq(elementtype, feat, ns)
        seq = strip_modifications(seq)
        if len(seq) >= minlen and len(seq) <= maxlen:
            yield formatting.string_and_clear(feat, ns)
        else:
            formatting.clear_el(feat)


def strip_modifications(seq):
    return re.sub('\[UNIMOD:\d*\]', '', seq)


def filter_whole_proteins(elements, protein_fasta, lookup, seqtype, ns,
                          deamidation, minpeplen, enforce_tryp):
    whole_proteins = {str(prot.seq).replace('L', 'I'): prot.id for prot in
                      fasta.parse_fasta(protein_fasta)}
    whole_proteins = {v: k for k, v in whole_proteins.items()}
    for element in elements:
        seq_matches_protein = False
        element_seqs = get_seqs_from_element(element, seqtype, ns, deamidation)
        element_prots = {seq: [(protid, pos) for protid, pos in
                               lookup.get_protein_from_pep(seq[:minpeplen])]
                         for seq in element_seqs}
        for pepseq, proteins in element_prots.items():
            for prot_id, pos in proteins:
                protseq = whole_proteins[prot_id]
                if pepseq in protseq:
                    if enforce_tryp and (pos == 0 or not set(
                            [pepseq[-1],
                             protseq[pos - 1]]).difference(['K', 'R'])):
                        # pepseq is tryptic on both ends, or
                        # pepseq is an N-term peptide),
                        # matches to protein seq so remove
                        seq_matches_protein = True
                        break
                    elif not enforce_tryp:
                        seq_matches_protein = True
                        break
        if seq_matches_protein:
            formatting.clear_el(element)
        else:
            yield formatting.string_and_clear(element, ns)


def filter_known_searchspace(elements, seqtype, lookup, ns, ntermwildcards,
                             deamidation):
    """Yields peptides from generator as long as their sequence is not found in
    known search space dict. Useful for excluding peptides that are found in
    e.g. ENSEMBL or similar"""
    for element in elements:
        seq_is_known = False
        for seq in get_seqs_from_element(element, seqtype, ns, deamidation):
            if lookup.check_seq_exists(seq, ntermwildcards):
                seq_is_known = True
                break
        if seq_is_known:
            formatting.clear_el(element)
        else:
            yield formatting.string_and_clear(element, ns)


def get_seqs_from_element(element, seqtype, ns, deamidation):
        seq = get_either_seq(seqtype, element, ns)
        seq = strip_modifications(seq)
        # Exchange leucines for isoleucines since MS can't differ and we
        # don't want to find 'novel' peptides which only have a difference
        # in this amino acid
        seq = seq.replace('L', 'I')
        if deamidation:
            return combination_replace(seq, 'D', 'N')
        else:
            return [seq]


def combination_replace(seq, from_aa, to_aa):
    options = [(c,) if c != from_aa else (from_aa, to_aa) for c in seq]
    return list(''.join(o) for o in product(*options))


def filter_unique_peptides(peptides, score, ns):
    """ Filters unique peptides from multiple Percolator output XML files.
        Takes a dir with a set of XMLs, a score to filter on and a namespace.
        Outputs an ElementTree.
    """
    scores = {'q': 'q_value',
              'pep': 'pep',
              'p': 'p_value',
              'svm': 'svm_score'}
    highest = {}
    for el in peptides:
        featscore = float(el.xpath('xmlns:%s' % scores[score],
                                   namespaces=ns)[0].text)
        seq = reader.get_peptide_seq(el, ns)

        if seq not in highest:
            highest[seq] = {
                'pep_el': formatting.stringify_strip_namespace_declaration(
                    el, ns), 'score': featscore}
        if score == 'svm':  # greater than score is accepted
            if featscore > highest[seq]['score']:
                highest[seq] = {
                    'pep_el':
                    formatting.stringify_strip_namespace_declaration(el, ns),
                    'score': featscore}
        else:  # lower than score is accepted
            if featscore < highest[seq]['score']:
                highest[seq] = {
                    'pep_el':
                    formatting.stringify_strip_namespace_declaration(el, ns),
                    'score': featscore}
        formatting.clear_el(el)

    for pep in list(highest.values()):
        yield pep['pep_el']



from app.writers import tsv


def write_mzid_tsv(header, psms, outfn):
    tsv.write_tsv(header, psms, outfn)


def write_multi_mzidtsv(header, oldheader, psms, base_outfile):
    outfile_handles = {}
    for psm in psms:
        line = psm['psm']
        split_pool = psm['split_pool']
        if split_pool not in outfile_handles:
            outfile_handles[split_pool] = open(base_outfile.format(split_pool),
                                               'w')
            tsv.write_tsv_line_from_list(header, outfile_handles[split_pool])
        tsv.write_tsv_line_from_list([line[field] for field in oldheader],
                                     outfile_handles[split_pool])
    [handle.close() for handle in outfile_handles.values()]



from app.actions.mzidtsv import quant as prep
from app.drivers.mzidtsv import MzidTSVDriver
from app.drivers.options import mzidtsv_options


class TSVQuantDriver(MzidTSVDriver):
    lookuptype = 'quant'
    outsuffix = '_quant.tsv'
    command = 'quant'
    commandhelp = ('Add quantitative data to a tab separated file with'
                   'PSMs. Quant data is fetched from an SQLite lookup '
                   'Specify --isobaric and/or --precursor.')

    def set_options(self):
        super().set_options()
        options = self.define_options(['lookupfn', 'precursor', 'isobaric'],
                                      mzidtsv_options)
        self.options.update(options)

    def get_psms(self):
        """Creates iterator to write to new tsv. Contains input tsv
        lines plus quant data for these."""
        self.header, isob_header = prep.get_full_and_isobaric_headers(
            self.oldheader, self.lookup, self.isobaric, self.precursor)
        self.psms = prep.generate_psms_quanted(self.lookup, self.fn,
                                               isob_header, self.oldheader,
                                               self.isobaric, self.precursor)



from Bio import SeqIO


def get_proteins_for_db(fastafn):
    """Runs through fasta file and returns proteins accession nrs, sequences
    and evidence levels for storage in lookup DB. Duplicate accessions in
    fasta are accepted and removed by keeping only the last one.
    """
    objects = {}
    for record in parse_fasta(fastafn):
        objects[parse_protein_identifier(record)] = record
    return (((acc,) for acc in list(objects)),
            ((acc, str(record.seq)) for acc, record in objects.items()),
            ((acc, get_uniprot_evidence_level(record.description))
             for acc, record in objects.items()))


def parse_biomart_fn(martfn, ensg_id, ensp_id, desc_id, symb_id, alt_symb_id):
    with open(martfn) as fp:
        header = next(fp).strip().split('\t')
        ensg = header.index(ensg_id)
        ensp = header.index(ensp_id)
        desc = header.index(desc_id)
        try:
            symb = header.index(symb_id)
        except ValueError:
            symb = header.index(alt_symb_id)
        for line in fp:
            line = line.strip('\n').split('\t')
            if line == ['']:
                continue
            protein = line[ensp]
            if not protein:
                continue
            yield (protein, line[ensg], line[symb], line[desc])


def get_proteins_genes(fastafn, fastadelim=None, genefield=None):
    """This returns a tuple of (protein, gene, HGNC symbol, description) from
    a passed file. If the file is FASTA from ENSEMBL or UniProt, only genes and
    descriptions are given and symbol will be None. If the file is a ENSEMBL
    Biomart mapping file it tries to parse that and also return the rest"""
    with open(fastafn) as fp:
        firstline = next(fp).strip()
    if firstline[0] == '>':
        for record in parse_fasta(fastafn):
            rectype = get_record_type(record)
            yield (record.id, get_gene(record.description, rectype,
                                       fastadelim, genefield),
                   None, record.description)
    elif 'Ensembl Gene ID' in firstline.split('\t'):
        for line in parse_biomart_fn(fastafn, 'Ensembl Gene ID',
                                     'Ensembl Protein ID', 'Description',
                                     'HGNC symbol', 'Associated Gene Name'):
            yield line
    elif 'Gene ID' in firstline.split('\t'):
        for line in parse_biomart_fn(fastafn, 'Gene ID', 'Protein ID',
                                     'Description', 'HGNC symbol',
                                     'Associated Gene Name'):
            yield line


def parse_protein_identifier(record):
    return record.id


def parse_fasta(fn):
    with open(fn) as fp:
        for record in SeqIO.parse(fp, 'fasta'):
            yield record


def get_record_type(record):
    dmod = get_decoy_mod_string(record.id)
    test_name = record.id
    if dmod is not None:
        test_name = record.id.replace(dmod, '')
    if test_name.split('|')[0] in ['sp', 'tr']:
        return 'swiss'
    elif test_name[:3] == 'ENS':
        return 'ensembl'
    else:
        return False


def get_decoy_mod_string(protein):
    mods = ['tryp_reverse', 'reverse', 'decoy', 'random', 'shuffle']
    for mod in mods:
        if mod in protein:
            if protein.endswith('_{}'.format(mod)):
                return '_{}'.format(mod)
            elif protein.endswith('{}'.format(mod)):
                return mod
            elif protein.startswith('{}_'.format(mod)):
                return '{}_'.format(mod)
            elif protein.startswith('{}'.format(mod)):
                return mod


def get_gene(description, rectype, fastadelim, genefield):
    if not rectype:
        if None not in [fastadelim, genefield]:
            return description.split(fastadelim)[genefield]
        else:
            return None
    splitter = {'ensembl': ':',
                'swiss': '='}[rectype]
    field = {'ensembl': 'gene',
             'swiss': 'GN'}[rectype]
    splitdesc = [x.split(splitter) for x in description.split()
                 if splitter in x]
    gn_list = [x[1] for x in splitdesc if x[0] == field]
    try:
        return gn_list[0]
    except IndexError:
        return 'NA'


def get_uniprot_evidence_level(header):
    """Returns uniprot protein existence evidence level for a fasta header.
    Evidence levels are 1-5, but we return 5 - x since sorting still demands
    that higher is better."""
    header = header.split()
    for item in header:
        item = item.split('=')
        try:
            if item[0] == 'PE':
                return 5 - int(item[1])
        except IndexError:
            continue
    return -1



from app.writers import prottable as writers
from app.drivers.base import BaseDriver


class PepProttableDriver(BaseDriver):
    """Base class for prottable.py"""
    def __init__(self):
        super().__init__()
        self.oldheader = False
        self.probability = False
        self.poolnames = False
        self.group_by_field = False

    def run(self):
        self.initialize_input()
        self.create_header()
        self.set_feature_generator()
        self.write()
        self.finish()

    def write(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        writers.write_prottable(self.header, self.features, outfn)



from app.readers import tsv as reader
from app.writers import prottable as writers
from app.actions.headers import prottable as head
from app.drivers.pepprottable import PepProttableDriver


class ProttableDriver(PepProttableDriver):
    def __init__(self):
        self.genecentric = False
        super().__init__()
        self.infiletype = 'protein table'

    def create_header(self):
        self.headerfields = head.get_prottable_headerfields(self.headertypes,
                                                            self.lookup,
                                                            self.poolnames,
                                                            self.genecentric)
        self.header = head.generate_protein_header(self.headerfields, self.oldheader, 
                self.group_by_field, self.genecentric)


class ProttableAddData(ProttableDriver):
    def initialize_input(self):
        self.oldheader = reader.get_tsv_header(self.fn)
        self.in_proteins = reader.generate_tsv_proteins(self.fn,
                                                        self.oldheader)


class ProttableMergeDriver(ProttableDriver):
    def __init__(self):
        super().__init__()
        self.group_by_field = True

    def initialize_input(self):
        self.headertypes = ['proteindata']
        for inflag, htype in zip([self.probability, self.fdr,
                                  self.pep, self.precursor,
                                  self.isobaric],
                                 ['probability', 'proteinfdr',
                                  'proteinpep', 'precursorquant', 'isoquant']):
            if inflag:
                self.headertypes.append(htype)
        self.poolnames = [x[0] for x in self.lookup.get_all_poolnames()]

    def write(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        writers.write_prottable_with_na(self.header, self.features, outfn)



from app.lookups.sqlite.base import ResultLookupInterface


class QuantDB(ResultLookupInterface):

    def select_all_psm_quants(self, isobaric=False, precursor=False):
        selects = ['pr.rownr']
        joins = ['JOIN psms USING(psm_id)', 'JOIN mzml USING(spectra_id)']
        sqlfields, fieldcount = {}, 1
        if isobaric:
            selects.extend(['ic.channel_name', 'iq.intensity'])
            joins.extend(['JOIN isobaric_quant AS iq USING(spectra_id)',
                          'JOIN isobaric_channels AS ic USING(channel_id)'])
            sqlfields['isochan'] = fieldcount
            sqlfields['isoquant'] = fieldcount + 1
            fieldcount += 2
        if precursor:
            selects.extend(['pq.intensity'])
            joins.extend(['LEFT OUTER JOIN ms1_align USING(spectra_id)',
                          'LEFT OUTER JOIN ms1_quant AS pq USING(feature_id)'])
            sqlfields['precursor'] = fieldcount
        if not precursor and not isobaric:
            raise RuntimeError('Cannot add quantification data, neither '
                               'isobaric, nor precursor have been specified.')
        sql = ('SELECT {} FROM psmrows as pr {} '
               'ORDER BY pr.rownr'.format(', '.join(selects), ' '.join(joins)))
        cursor = self.get_cursor()
        return cursor.execute(sql), sqlfields

    def get_precursor_quant_window(self, windowsize, minmz):
        cursor = self.get_cursor()
        return cursor.execute(
            'SELECT feature_id, mzmlfile_id, charge, mz, retention_time '
            'FROM ms1_quant '
            'WHERE mz > ? ORDER BY mz '
            'LIMIT ?', (minmz, windowsize))

    def get_all_quantmaps(self):
        """Returns all unique quant channels from lookup as list"""
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT channel_name FROM isobaric_channels')
        return cursor.fetchall()


class IsobaricQuantDB(QuantDB):
    def add_tables(self):
        self.create_tables(['isobaric_quant', 'isobaric_channels'])

    def store_channelmap(self, channels):
        self.store_many(
            'INSERT INTO isobaric_channels(channel_name) VALUES(?)', channels)

    def store_isobaric_quants(self, quants):
        self.store_many(
            'INSERT INTO isobaric_quant(spectra_id, channel_id, intensity) '
            'VALUES (?, ?, ?)', quants)

    def index_isobaric_quants(self):
        self.index_column('spectraid_index', 'isobaric_quant', 'spectra_id')
        self.index_column('channel_id_index', 'isobaric_quant', 'channel_id')

    def get_channelmap(self):
        cursor = self.get_cursor()
        cursor.execute('SELECT channel_id, channel_name FROM isobaric_channels')
        return cursor


class PrecursorQuantDB(QuantDB):
    def add_tables(self):
        self.create_tables(['ms1_quant', 'ms1_align'])

    def store_ms1_quants(self, quants):
        self.store_many(
            'INSERT INTO ms1_quant(mzmlfile_id, retention_time, mz, '
            'charge, intensity) VALUES (?, ?, ?, ?, ?)', quants)

    def store_ms1_alignments(self, aligns):
        self.store_many(
            'INSERT INTO ms1_align(spectra_id, feature_id) '
            'VALUES (?, ?)', aligns)

    def index_precursor_quants(self):
        self.index_column('charge_index', 'ms1_quant', 'charge')
        self.index_column('feat_mz_index', 'ms1_quant', 'mz')

    def get_spectra_mz_sorted(self):
        return self.get_cursor().execute(
            'SELECT spectra_id, mzmlfile_id, charge, mz, retention_time '
            'FROM mzml ORDER BY mz')



from app.drivers.pycolator import base
from app.actions.pycolator import reassign
from app.drivers.options import pycolator_options


class ReassignmentDriver(base.PycolatorDriver):
    """Reassigns statistics from qvality output on a percolator output file"""
    outsuffix = '_reassigned.xml'
    command = 'reassign'
    commandhelp = ('Reassigns statistics from a qvality output file onto a '
                   'single percolator input file.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['qvalityout', 'featuretype'],
                                                pycolator_options))

    def set_features(self):
        stats = reassign.parse_qvality_output(self.qvalityout)
        if self.featuretype == 'psm':
            self.features = {
                'psm': reassign.reassign_elements(self.allpsms, stats,
                                                  self.ns),
                'peptide': self.get_all_peptides_strings(),
            }
        elif self.featuretype == 'peptide':
            self.features = {
                'peptide': reassign.reassign_elements(self.allpeps, stats,
                                                      self.ns),
                'psm': self.get_all_psms_strings(),
            }



from app.readers import tsv as tsvreader
from app.dataformats import mzidtsv as mzidtsvdata
from app.lookups.sqlite import proteingroups as lookups
from app.actions.mzidtsv import proteingroup_sorters as sorters


def get_header_with_proteingroups(header):
    ix = header.index(mzidtsvdata.HEADER_PROTEIN) + 1
    return header[:ix] + mzidtsvdata.HEADER_PG + header[ix:]


def get_all_proteins_from_unrolled_psm(rownr, pgdb):
    return pgdb.get_proteins_for_peptide([rownr])


def generate_psms_with_proteingroups(fn, oldheader, newheader, pgdb,
                                     unroll=False):
    rownr = 0
    use_evi = pgdb.check_evidence_tables()
    all_protein_group_content = pgdb.get_all_psms_proteingroups(use_evi)
    protein = next(all_protein_group_content)
    for psm in tsvreader.generate_tsv_psms(fn, oldheader):
        if unroll:
            psm_id = tsvreader.get_psm_id(psm)
            lineproteins = get_all_proteins_from_unrolled_psm(psm_id, pgdb)
        else:
            lineproteins = tsvreader.get_proteins_from_psm(psm)
        proteins_in_groups = {}
        while protein[0] == rownr:
            try:
                proteins_in_groups[protein[
                    lookups.MASTER_INDEX]].append(protein)
            except KeyError:
                proteins_in_groups[protein[lookups.MASTER_INDEX]] = [protein]
            try:
                protein = next(all_protein_group_content)
            except StopIteration:
                protein = [-1]
        sorted_pgs = sorters.sort_protein_groups(proteins_in_groups, use_evi)
        psm_masters = []
        psm_pg_proteins = []
        for master, group in sorted_pgs.items():
            psm_masters.append(master)
            psm_pg_proteins.append([protein[lookups.PROTEIN_ACC_INDEX]
                                    for protein in group])
        outpsm = {mzidtsvdata.HEADER_MASTER_PROT: ';'.join(psm_masters),
                  mzidtsvdata.HEADER_PG_CONTENT: ';'.join(
                      [','.join([y for y in x]) for x in psm_pg_proteins]),
                  mzidtsvdata.HEADER_PG_AMOUNT_PROTEIN_HITS: ';'.join(
                      count_protein_group_hits(lineproteins, psm_pg_proteins))
                  }
        outpsm.update(psm)
        rownr += 1
        yield outpsm


def count_protein_group_hits(lineproteins, groups):
    """Takes a list of protein accessions and a list of protein groups
    content from DB. Counts for each group in list how many proteins
    are found in lineproteins. Returns list of str amounts.
    """
    hits = []
    for group in groups:
        hits.append(0)
        for protein in lineproteins:
            if protein in group:
                hits[-1] += 1
    return [str(x) for x in hits]



import os
from app.readers import xml as basereader
from app.readers import ml
from app.readers import xmlformatting as formatting


def mzmlfn_ms2_spectra_generator(mzmlfiles):
    for fn, spec, ns in mzmlfn_spectra_generator(mzmlfiles):
        specparams = get_all_cvparams(spec, ns)
        mslvl = fetch_cvparam_value_by_name(specparams, 'ms level')
        if mslvl != '2':
            continue
        scannr = get_spec_scan_nr(spec)
        rt = fetch_cvparams_values_from_subel(spec, 'scan',
                                              ['scan start time'], ns)
        iit = fetch_cvparams_values_from_subel(spec, 'scan',
                                               ['ion injection time'], ns)
        mz, charge = fetch_cvparams_values_from_subel(spec, 'selectedIon',
                                                      ['selected ion m/z',
                                                       'charge state'], ns)
        yield fn, {'scan': scannr, 'rt': rt[0], 'iit': iit[0], 'mz': mz,
                   'charge': charge}
        formatting.clear_el(spec)


def mzmlfn_spectra_generator(mzmlfiles):
    for fn in mzmlfiles:
        ns = basereader.get_namespace(fn)
        spectra = basereader.generate_xmltags(
            fn,
            'spectrum',
            ['offset',
             ],
            ns)
        for spectrum in spectra:
            yield os.path.basename(fn), spectrum, ns


def fetch_cvparams_values_from_subel(base, subelname, paramnames, ns):
    """Searches a base element for subelement by name, then takes the
    cvParams of that subelement and returns the values as a list
    for the paramnames that match. Value order in list equals input
    paramnames order."""
    sub_el = basereader.find_element_xpath(base, subelname, ns)
    cvparams = get_all_cvparams(sub_el, ns)
    output = []
    for param in paramnames:
        output.append(fetch_cvparam_value_by_name(cvparams, param))
    return output


def fetch_cvparam_value_by_name(params, name):
    for cvparam in params:
        try:
            if cvparam.attrib['name'] == name:
                return cvparam.attrib['value']
        except KeyError:
            pass
    return False


def get_all_cvparams(element, ns):
    return element.findall('{%s}cvParam' % ns['xmlns'])


def get_spec_scan_nr(spectrum):
    """Returns scan number of mzML spectrum as a str."""
    return ml.get_scan_nr(spectrum, 'id')



from setuptools import setup, find_packages


###################################################################
NAME = 'msstitch'
PACKAGES = find_packages(where='src')
KEYWORDS = ['mass spectrometry', 'proteomics', 'processing']
CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Science/Research',
    'Natural Language :: English',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
]
INSTALL_REQUIRES = ['numpy', 'lxml', 'biopython', 'pyyaml']
METADATA = {
    'version': '2.12',
    'title': 'msstitch',
    'description': 'MS proteomics post processing utilities',
    'uri': 'https://github.com/glormph/msstitch',
    'author': 'Jorrit Boekel',
    'email': 'jorrit.boekel@scilifelab.se',
    'license': 'MIT',
    'copyright': 'Copyright (c) 2013 Jorrit Boekel',
}

CLI = {'console_scripts': ['msspercolator=app.pycolator:main',
                           'msslookup=app.mslookup:main',
                           'msspsmtable=app.mzidtsv:main',
                           'msspeptable=app.peptable:main',
                           'mssprottable=app.prottable:main',
                           ]}
###################################################################

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst', format='md')
except(IOError, ImportError):
    long_description = open('README.md').read()


if __name__ == '__main__':
    setup(
        name=NAME,
        description=METADATA['description'],
        license=METADATA['license'],
        url=METADATA['uri'],
        version=METADATA['version'],
        author=METADATA['author'],
        author_email=METADATA['email'],
        maintainer=METADATA['author'],
        maintainer_email=METADATA['email'],
        keywords=KEYWORDS,
        packages=PACKAGES,
        package_dir={'': 'src'},
        long_description=long_description,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        entry_points=CLI,
    )



from app.readers import tsv as reader
from app.actions.prottable import precursorarea as preparation
from app.drivers.prottable.base import ProttableAddData
from app.drivers.options import prottable_options


class AddPrecursorAreaDriver(ProttableAddData):
    outsuffix = '_ms1q.txt'
    command = 'ms1quant'
    commandhelp = ('Add MS1 quantification data from a PSM table '
                   'containing precursor quant areas.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['psmfile', 'proteincol'],
                                                prottable_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.headertypes = ['precursorquant']

    def initialize_input(self):
        super().initialize_input()
        pepheader = reader.get_tsv_header(self.psmfile)
        self.get_column_header_for_number(['proteincol'], pepheader)
        self.in_peptides = reader.generate_tsv_peptides(self.psmfile)

    def set_feature_generator(self):
        self.features = preparation.add_ms1_quant_from_top3_mzidtsv(
            self.in_proteins, self.in_peptides, self.headerfields,
            self.proteincol)



from collections import OrderedDict


from app.dataformats import peptable as peptabledata
from app.actions.mergetable import (simple_val_fetch, fill_mergefeature,
                                    get_isobaric_quant)
from app.actions.proteindata import create_featuredata_map


def build_peptidetable(pqdb, headerfields, isobaric=False,
                       precursor=False, fdr=False, pep=False,
                       genecentric=False):
    """Fetches peptides and quants from joined lookup table, loops through
    them and when all of a peptides quants/data have been collected, yields
    peptide quant information."""
    peptidedatamap = create_featuredata_map(pqdb, genecentric=genecentric, 
            psm_fill_fun=add_psm_to_peptidedata, pgene_fill_fun=add_protgene_to_pepdata,
            is_peptides=True)
    count_psms(peptidedatamap)
    empty_return = lambda x, y, z: {}
    iso_fun = {True: get_isobaric_quant, False: empty_return}[isobaric]
    ms1_fun = {True: get_precursor_quant, False: empty_return}[precursor]
    fdr_fun = {True: get_pep_fdr,
               False: empty_return}[fdr]
    pep_fun = {True: get_peptide_pep,
               False: empty_return}[pep]
    pdata_fun = get_protein_data
    peptide_sql, sqlfieldmap = pqdb.prepare_mergetable_sql(precursor, isobaric,
                                                           probability=False,
                                                           fdr=fdr, pep=pep)
    peptides = pqdb.get_merged_features(peptide_sql)
    peptide = next(peptides)
    outpeptide = {peptabledata.HEADER_PEPTIDE: peptide[sqlfieldmap['p_acc']]}
    check_pep = {k: v for k, v in outpeptide.items()}
    fill_mergefeature(outpeptide, iso_fun, ms1_fun, empty_return, fdr_fun,
                      pep_fun, pdata_fun, peptide, sqlfieldmap,
                      headerfields, peptidedatamap)
    for peptide in peptides:
        p_seq = peptide[sqlfieldmap['p_acc']]
        if p_seq != outpeptide[peptabledata.HEADER_PEPTIDE]:
            if outpeptide != check_pep:
                yield outpeptide
            outpeptide = {peptabledata.HEADER_PEPTIDE: p_seq}
            check_pep = {k: v for k, v in outpeptide.items()}
        fill_mergefeature(outpeptide, iso_fun, ms1_fun, empty_return, fdr_fun,
                          pep_fun, pdata_fun, peptide, sqlfieldmap,
                          headerfields, peptidedatamap)
    yield outpeptide


def get_precursor_quant(peptide, sqlmap, headerfields):
    return simple_val_fetch(peptide, sqlmap,
                            headerfields['precursorquant'][
                                peptabledata.HEADER_AREA], 'preq_val')


def get_pep_fdr(peptide, sqlmap, headerfields):
    return simple_val_fetch(peptide, sqlmap,
                            headerfields['peptidefdr'][
                                peptabledata.HEADER_QVAL], 'fdr_val')


def get_peptide_pep(peptide, sqlmap, headerfields):
    return simple_val_fetch(peptide, sqlmap,
                            headerfields['peptidepep'][
                                peptabledata.HEADER_PEP], 'pep_val')


def get_no_psms(peptide, pdata, headerfields):
    hfields = [peptabledata.HEADER_NO_PSM,
               ]
    seq = peptide[peptabledata.HEADER_PEPTIDE]
    outdict = {}
    for pool, psms in pdata[seq]['psms'].items():
        pool_values = [psms]
        outdict.update({headerfields['nopsms'][hfield][pool]: val
                        for (hfield, val) in zip(hfields, pool_values)})
    return outdict


def get_protein_data(peptide, pdata, headerfields, accfield):
    """These fields are currently not pool dependent so headerfields
    is ignored"""
    report = get_proteins(peptide, pdata, headerfields)
    return get_cov_descriptions(peptide, pdata, report)


def get_proteins(peptide, pdata, headerfields):
    seq = peptide[peptabledata.HEADER_PEPTIDE]
    outdict = {}
    try:
        proteins = ';'.join([x[0] for x in pdata[seq]['proteins']])
    except TypeError:
        pass
    else:
        outdict = {peptabledata.HEADER_PROTEINS: proteins}
    for pool, psms in pdata[seq]['psms'].items():
        outdict.update(
            {headerfields['proteindata'][peptabledata.HEADER_NO_PSM][pool]:
             psms})
    return outdict


def get_cov_descriptions(peptide, pdata, report):
    def format_val(value, valtype):
        formatter = {int: lambda x: str(x),
                     float: lambda x: str(float(x)),
                     str: lambda x: x,
                     }
        return formatter[valtype](value)

    seq = peptide[peptabledata.HEADER_PEPTIDE]
    for idx, key, keytype in zip([1, 2, 3, 4, 5],
                                 [peptabledata.HEADER_COVERAGES,
                                  peptabledata.HEADER_DESCRIPTIONS,
                                  peptabledata.HEADER_GENES,
                                  peptabledata.HEADER_ASSOCIATED,
                                  peptabledata.HEADER_NO_CONTENTPROTEINS],
                                 [float, str, str, str, int]):
        try:
            replist = [format_val(x[idx], keytype)
                       for x in pdata[seq]['proteins']]
        except (TypeError, IndexError):
            # index too high, or None skip, item will be NA'ed downstream
            continue
        if key in [peptabledata.HEADER_GENES, peptabledata.HEADER_ASSOCIATED]:
            replist = OrderedDict([(x, 1) for x in replist]).keys()
        try:
            report[key] = ';'.join(replist)
        except TypeError:
            # None in list, skip, NA will be output
            continue
    return report


def count_psms(pdata):
    for seq in pdata:
        for pool in pdata[seq]['psms']:
            pdata[seq]['psms'][pool] = len(pdata[seq]['psms'][pool])


def add_protgene_to_pepdata(peptidedata, seq, p_acc, dbrec, genecentric, pgcontentmap=None):
    if genecentric == 'plain':
        gene, desc, assoc_id, cov, pgcontent = None, None, None, None, None
        protein = (p_acc, cov, desc, gene, assoc_id)
    elif genecentric:
        gene = p_acc
        desc, assoc_id = dbrec[2], dbrec[3]
        cov, pgcontent = None, None
        protein = (None, cov, desc, gene, assoc_id)
    else:
        desc, cov = dbrec[2], dbrec[3]
        gene, assoc_id = dbrec[4], dbrec[5]
        pgcontent = pgcontentmap[p_acc]
        protein = (p_acc, cov, desc, gene, assoc_id, len(pgcontent))
    try:
        peptidedata[seq]['proteins'].add(protein)
    except KeyError:
        peptidedata[seq] = {'proteins': set()}
        peptidedata[seq]['proteins'].add(protein)


def add_psm_to_peptidedata(peptidedata, p_acc, pool, psmdata):
    seq, psm_id = psmdata[2], psmdata[3]
    try:
        peptidedata[seq]['psms'][pool].add(psm_id)
    except KeyError:
        try:
            peptidedata[seq]['psms'][pool] = set()
        except KeyError:
            peptidedata[seq].update({'psms': {pool: set()}})
        peptidedata[seq]['psms'][pool].add(psm_id)



from app.drivers.mslookup import base
from app.readers import openms as openmsreader
from app.readers import tsv as tsvreader
from app.actions.mslookup import quant as lookups
from app.drivers.options import mslookup_options


class QuantLookupDriver(base.LookupDriver):
    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['spectrafns', 'multifiles'],
                                                mslookup_options))


class IsobaricQuantLookupDriver(QuantLookupDriver):
    lookuptype = 'isoquant'
    command = 'isoquant'
    commandhelp = ('Create lookup of isobaric quant data in OpenMS '
                   'consensusXML format. Use requires --spectra, --dbfile '
                   'with an sqlite lookup of spectra, and passing multiple '
                   'consensusXML files to -i.')

    def __init__(self):
        super().__init__()
        self.infiletype = 'consensusXML'

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.cons_fns = self.fn

    def create_lookup(self):
        quantmap = openmsreader.get_quantmap(self.cons_fns[0])
        mzmlfn_consxml = openmsreader.mzmlfn_cons_el_generator(self.spectrafns,
                                                               self.cons_fns)
        lookups.create_isobaric_quant_lookup(self.lookup, mzmlfn_consxml,
                                             quantmap),


class PrecursorQuantLookupDriver(QuantLookupDriver):
    lookuptype = 'ms1quant'
    command = 'ms1quant'
    commandhelp = ('Creates lookup of precursor quant data in OpenMS '
                   'featureXML, or Kronik output formats. '
                   'Use requires --spectra, --dbfile with an sqlite lookup of '
                   'spectra, --quanttype to determine quant output, --mztol, '
                   '--rttol, --mztoltype for tolerance specification, and '
                   'passing a featureXML or kronik file to -i.')

    def __init__(self):
        super().__init__()
        self.infiletype = 'featureXML or Kronik'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['quantfiletype', 'rttol',
                                                 'mztol', 'mztoltype'],
                                                mslookup_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.precursorfns = self.fn

    def create_lookup(self):
        if self.quantfiletype == 'openms':
            specfn_feats = openmsreader.mzmlfn_feature_generator(
                self.spectrafns, self.precursorfns)
        elif self.quantfiletype == 'kronik':
            specfn_feats = tsvreader.mzmlfn_kronikfeature_generator(
                self.spectrafns, self.precursorfns)
        lookups.create_precursor_quant_lookup(self.lookup, specfn_feats,
                                              self.quantfiletype,
                                              self.rt_tol,
                                              self.mz_tol,
                                              self.mz_toltype)



from decimal import Decimal

from app.readers import openms as openmsreader
DB_STORE_CHUNK = 500000
FEATURE_ALIGN_WINDOW_AMOUNT = 1000
PROTON_MASS = 1.0072


def create_isobaric_quant_lookup(quantdb, specfn_consensus_els, channelmap):
    """Creates an sqlite lookup table of scannrs with quant data.

    spectra - an iterable of tupled (filename, spectra)
    consensus_els - a iterable with consensusElements"""
    # store quantchannels in lookup and generate a db_id vs channel map
    channels_store = ((name,) for name, c_id
                      in sorted(channelmap.items(), key=lambda x: x[1]))
    quantdb.store_channelmap(channels_store)
    channelmap_dbid = {channelmap[ch_name]: ch_id for ch_id, ch_name in
                       quantdb.get_channelmap()}
    quants = []
    mzmlmap = quantdb.get_mzmlfile_map()
    for specfn, consensus_el in specfn_consensus_els:
        rt = openmsreader.get_consxml_rt(consensus_el)
        rt = round(float(Decimal(rt) / 60), 12)
        qdata = get_quant_data(consensus_el)
        spectra_id = quantdb.get_spectra_id(mzmlmap[specfn],
                                            retention_time=rt)
        for channel_no in sorted(qdata.keys()):
            quants.append((spectra_id, channelmap_dbid[channel_no],
                           qdata[channel_no]))
            if len(quants) == DB_STORE_CHUNK:
                quantdb.store_isobaric_quants(quants)
    quantdb.store_isobaric_quants(quants)
    quantdb.index_isobaric_quants()


def create_precursor_quant_lookup(quantdb, mzmlfn_feats, quanttype,
                                  rttol, mztol, mztoltype):
    """Fills quant sqlite with precursor quant from:
        features - generator of xml features from openms
    """
    featparsermap = {'kronik': kronik_featparser,
                     'openms': openms_featparser,
                     }
    features = []
    mzmlmap = quantdb.get_mzmlfile_map()
    for specfn, feat_element in mzmlfn_feats:
        feat = featparsermap[quanttype](feat_element)
        features.append((mzmlmap[specfn], feat['rt'], feat['mz'],
                         feat['charge'], feat['intensity'])
                        )
        if len(features) == DB_STORE_CHUNK:
            quantdb.store_ms1_quants(features)
            features = []
    quantdb.store_ms1_quants(features)
    quantdb.index_precursor_quants()
    align_quants_psms(quantdb, rttol, mztol, mztoltype)


def get_minmax(center, tolerance, toltype=None):
    center = float(center)
    if toltype == 'ppm':
        tolerance = int(tolerance) / 1000000 * center
    else:
        tolerance = float(tolerance)
    return center - tolerance, center + tolerance


def align_quants_psms(quantdb, rt_tolerance, mz_tolerance, mz_toltype):
    allspectra = quantdb.get_spectra_mz_sorted()
    featwindow_max_mz = -1
    spec_feat_store = []
    for spec_id, fn_id, charge, mz, rt in allspectra:
        minmz, maxmz = get_minmax(mz, mz_tolerance, mz_toltype)
        if maxmz > featwindow_max_mz:
            feat_map, featwindow_max_mz = get_precursors_from_window(quantdb,
                                                                     minmz)
        best_feat_id = align_psm(mz, rt, fn_id, charge, feat_map, rt_tolerance)
        if not best_feat_id:
            continue
        spec_feat_store.append((spec_id, best_feat_id))
        if len(spec_feat_store) > DB_STORE_CHUNK:
            quantdb.store_ms1_alignments(spec_feat_store)
            spec_feat_store = []
    quantdb.store_ms1_alignments(spec_feat_store)


def align_psm(psm_mz, psm_rt, fn_id, charge, featmap, rttol):
    minrt, maxrt = get_minmax(psm_rt, rttol / 60)
    alignments = {}
    try:
        featlist = featmap[fn_id][charge]
    except KeyError:
        return False
    for feat_mz, feat_rt, feat_id in featlist:
        if abs(psm_rt - feat_rt) > rttol:
            continue
        alignments[abs(psm_mz - feat_mz)] = feat_id
    try:
        return alignments[min(alignments)]
    except ValueError:
        return False


def get_precursors_from_window(quantdb, minmz):
    """Returns a dict of a specified amount of features from the
    ms1 quant database, and the highest mz of those features"""
    featmap = {}
    mz = False
    features = quantdb.get_precursor_quant_window(FEATURE_ALIGN_WINDOW_AMOUNT,
                                                  minmz)
    for feat_id, fn_id, charge, mz, rt in features:
        try:
            featmap[fn_id][charge].append((mz, rt, feat_id))
        except KeyError:
            try:
                featmap[fn_id][charge] = [(mz, rt, feat_id)]
            except KeyError:
                featmap[fn_id] = {charge: [(mz, rt, feat_id)]}
    return featmap, mz


def kronik_featparser(feature):
    charge = int(feature['Charge'])
    mz = (float(feature['Monoisotopic Mass']) + charge * PROTON_MASS) / charge
    return {'rt': round(float(feature['Best RTime']), 12),
            'mz': mz,
            'charge': charge,
            'intensity': float(feature['Best Intensity']),
            }


def openms_featparser(feature):
    feat = openmsreader.get_feature_info(feature)
    feat['rt'] = round(float(Decimal(feat['rt']) / 60), 12)
    return feat


def get_quant_data(cons_el):
    """Gets quant data from consensusXML element"""
    quant_out = {}
    for reporter in cons_el.findall('.//element'):
        quant_out[reporter.attrib['map']] = reporter.attrib['it']
    return quant_out



from app.drivers.prottable.base import ProttableAddData
from app.actions.prottable import fdr as action
from app.readers import tsv as reader
from app.drivers.options import prottable_options


class ProttableFDRDriver(ProttableAddData):
    """Runs qvality on two TSV tables"""
    outsuffix = '_protfdr.txt'
    command = 'protfdr'
    commandhelp = ('Calculate FDR of protein (or tsv) tables '
                   'containing target (-i) proteins and decoy (--decoy) '
                   'proteins. Uses Q-score from Savitski 2014 MCP as a '
                   'ranking mechanism, q-values are #D/(#D+#T) without '
                   'correction')

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.headertypes = ['proteinfdr']

    def initialize_input(self):
        self.target = self.fn
        self.oldheader = reader.get_tsv_header(self.target)
        self.targetheader = reader.get_tsv_header(self.target)
        self.decoyheader = reader.get_tsv_header(self.decoyfn)

    def set_options(self):
        super().set_options()
        options = self.define_options(['decoyfn'], prottable_options)
        self.options.update(options)

    def prepare(self):
        """No percolator XML for protein tables"""
        self.target = self.fn
        self.targetheader = reader.get_tsv_header(self.target)
        self.decoyheader = reader.get_tsv_header(self.decoyfn)

    def set_feature_generator(self):
        self.features = action.generate_protein_fdr(
            self.target, self.decoyfn, self.targetheader, self.decoyheader,
            self.headerfields)


class PickedFDRDriver(ProttableFDRDriver):
    outsuffix = '_pickedfdr.txt'
    command = 'pickedfdr'
    commandhelp = ('Calculate FDR on gene tables containing target (-i) '
                   'proteins and decoy (--decoy) genes. Targets and '
                   'decoys will be filtered according to Savitski et al. '
                   '2014 MCP, where the best scoring of a pair of '
                   'target/decoy genes is retained and the other '
                   'gene discarded. Matching (reversed, tryptic reversed,'
                   ' scrambled) target and decoy FASTA files are needed to '
                   'determine the pairs, use --targetfasta, --decoyfasta. '
                   'Specify gene type with --picktype to distinguish between '
                   'creating target-decoy pairs from '
                   'matched FASTA files or from the gene table input. '
                   'Uses Q-score as a ranking mechanism, q-values are '
                   '#D/(#D+#T) without correction')

    def set_options(self):
        super().set_options()
        options = self.define_options(['t_fasta', 'd_fasta', 'picktype',
                                       'genefield', 'fastadelim'],
                                      prottable_options)
        self.options.update(options)

    def set_feature_generator(self):
        fastadelim, genefield = self.get_fastadelim_genefield(self.fastadelim,
                                                              self.genefield)
        self.features = action.generate_pick_fdr(
            self.target, self.decoyfn, self.targetheader, self.decoyheader,
            self.headerfields, self.t_fasta, self.d_fasta, self.picktype,
            fastadelim, genefield)


class ProttableAddFDRDriver(ProttableAddData):
    """Assigns FDR to protein table based on qvality output and protein table
    probabilities"""
    outsuffix = '_protfdr.txt'
    command = 'fdr'
    commandhelp = ('Add protein FDR to protein table by comparing '
                   'score (peptide q-value, protein probability, etc) '
                   'with qvality lookup table. Needs --scorecolpattern and '
                   'qvality output file specified with --qvality')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['qvalityout',
                                                 'scorecolpattern'],
                                                prottable_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.headertypes = ['proteinfdr', 'proteinpep']

    def initialize_input(self):
        super().initialize_input()
        self.scorecol = reader.get_cols_in_file(self.scorecolpattern,
                                                self.oldheader, True)

    def set_feature_generator(self):
        self.features = action.add_protein_fdr(self.qvalityout,
                                               self.in_proteins,
                                               self.headerfields,
                                               self.scorecol)



HEADER_ACCESSION = 'Accession'
HEADER_PROTEIN = 'Protein ID'
HEADER_PROTEINS = 'Protein ID(s)'
HEADER_GENE = 'Gene'
HEADER_GENEID = 'Gene ID'
HEADER_GENENAME = 'Gene Name'
HEADER_ASSOCIATED = 'Associated gene ID'
HEADER_DESCRIPTION = 'Description'
HEADER_COVERAGE = 'Coverage'
HEADER_NO_PROTEIN = '# Proteins'
HEADER_CONTENTPROT = 'Proteins in group'
HEADER_NO_UNIPEP = '# Unique peptides'
HEADER_NO_PEPTIDE = '# Peptides'
HEADER_NO_PSM = '# PSMs'
HEADER_NO_PSMS_SUFFIX = ' - # quanted PSMs'
HEADER_AREA = 'MS1 precursor area'
HEADER_NO_QUANT_PSM = '# Quantified PSMs'
HEADER_CV_QUANT_PSM = '# CV of quantified PSMs'
HEADER_PROBABILITY = 'Protein error probability'
HEADER_QVAL = 'q-value'
HEADER_PEP = 'PEP'
HEADER_QVAL_MODELED = 'q-value (linear model)'
HEADER_QSCORE = 'Q-score best peptide'
PICKED_HEADER = [HEADER_PROTEIN, HEADER_QSCORE]
ACCESSIONS = {
        False: HEADER_PROTEIN, 
        'genes': HEADER_GENEID,
        'assoc': HEADER_GENENAME,
        }



import sys

from app.readers import xmlformatting as formatting


def parse_qvality_output(fn):
    statistics = {}
    with open(fn) as fp:
        for line in fp:
            line = line.strip()
            line = line.split('\t')
            if len(line) == 3:
                try:
                    score = float(line[0])
                except ValueError:
                    continue
                statistics[score] = {'PEP': line[1], 'q': line[2]}
    return statistics


def reassign_elements(elements, stats, ns):
    for el in elements:
        score = round(float(el.xpath('xmlns:svm_score',
                                     namespaces=ns)[0].text), 5)
        oldq = el.xpath('xmlns:q_value', namespaces=ns)[0]
        oldpep = el.xpath('xmlns:pep', namespaces=ns)[0]
        newq, newpep, warning = lookup_statistic(score, stats)
        if warning is not None:
            sys.stdout.write(warning)
        oldq.text, oldpep.text = newq, newpep
        yield formatting.string_and_clear(el, ns)


def lookup_statistic(score, stats):
    """ Finds statistics that correspond to PSM/peptide/protein feature's
    score. Loops through list of qvality generated scores until it finds
    values closest to the feature's svm_score."""
    if score in stats:
        return stats[score]['q'], stats[score]['PEP'], None
    else:
        lower, warning = None, None
        for stat_score in sorted(stats.keys()):
            if score < stat_score:
                break
            lower = stat_score
        if score > stat_score:
            warning = ('WARNING! Values found with higher svm_score than in '
                       'qvality recalculation!')
        if not lower:
            warning = ('WARNING! Values found with lower svm_score than in '
                       'qvality recalculation were set to PEP=1, qval=1.')
            return '1', '1', warning
        qval = (float(stats[stat_score]['q']) + float(stats[lower]['q'])) / 2
        pep = (float(stats[stat_score]['PEP']) +
               float(stats[lower]['PEP'])) / 2
    return str(qval), str(pep), warning



from app.lookups.sqlite import (proteingroups, quant, searchspace,
                                biosets, spectra, prottable, psms,
                                peptable)


def get_lookup(fn, lookuptype):
    lookupmap = {'biosets': biosets.BioSetDB,
                 'spectra': spectra.SpectraDB,
                 'psm': psms.PSMDB,
                 'proteingroups': proteingroups.ProteinGroupDB,
                 'quant': quant.QuantDB,
                 'isoquant': quant.IsobaricQuantDB,
                 'ms1quant': quant.PrecursorQuantDB,
                 'searchspace': searchspace.SearchSpaceDB,
                 'peptidetable': peptable.PepTableProteinCentricDB,
                 'peptidegenecentrictable': peptable.PepTableGeneCentricDB,
                 'peptidetableplain': peptable.PepTablePlainDB,
                 'prottable': prottable.ProtTableDB,
                 'genetable': prottable.GeneTableDB,
                 'associdtable': prottable.GeneTableAssocIDsDB,
                 }
    return lookupmap[lookuptype](fn)


def create_new_lookup(fn, lookuptype):
    with open(fn, 'w'):
        pass
    return get_lookup(fn, lookuptype)



from collections import OrderedDict

from app.actions.headers.base import (generate_general_header,
                                      generate_headerfields,
                                      get_isoquant_fields)
from app.dataformats import prottable as prottabledata


def generate_protein_header(headerfields, oldheader, group_by_field, genecentric):
    """Returns a header as a list, ready to write to TSV file"""
    fieldtypes = ['proteindata', 'probability', 'proteinfdr', 'proteinpep',
                  'precursorquant', 'isoquant', 'bestpepscore']
    firstfield = prottabledata.ACCESSIONS[genecentric]
    return generate_general_header(headerfields, fieldtypes, firstfield, oldheader,
                                   group_by_field)


def get_prottable_headerfields(headertypes, lookup=False, poolnames=False, genecentric=False):
    """Called by driver to generate headerfields object"""
    field_defs = {'isoquant': get_isoquant_fields,
                  'precursorquant': get_precursorquant_fields,
                  'probability': get_probability_fields,
                  'proteindata': get_proteininfo_fields,
                  'proteinfdr': get_proteinfdr_fields,
                  'proteinpep': get_proteinpep_fields,
                  'bestpepscore': get_bestpeptide_fields,
                  }
    return generate_headerfields(headertypes, field_defs, poolnames, lookup, genecentric)


def get_precursorquant_fields(poolnames=False):
    return {prottabledata.HEADER_AREA: poolnames}


def get_probability_fields(poolnames=False):
    return {prottabledata.HEADER_PROBABILITY: poolnames}


def get_proteinfdr_fields(poolnames=False):
    return {prottabledata.HEADER_QVAL: poolnames}


def get_proteinpep_fields(poolnames=False):
    return {prottabledata.HEADER_PEP: poolnames}


def get_proteininfo_fields(poolnames=False, genecentric=False):
    """Returns header fields for protein (group) information.
    Some fields are shared between pools, others are specific
    for a pool"""
    allfields = OrderedDict()
    basefields = {
            False: [
                prottabledata.HEADER_GENEID, prottabledata.HEADER_GENENAME,
                prottabledata.HEADER_DESCRIPTION, prottabledata.HEADER_COVERAGE,
                prottabledata.HEADER_NO_PROTEIN, prottabledata.HEADER_CONTENTPROT,
                ],
            'genes': [
                prottabledata.HEADER_GENENAME, prottabledata.HEADER_PROTEINS,
                prottabledata.HEADER_DESCRIPTION],
            'assoc': [
                prottabledata.HEADER_GENEID, prottabledata.HEADER_PROTEINS,
                prottabledata.HEADER_DESCRIPTION],
            }[genecentric]
    poolfields = [prottabledata.HEADER_NO_UNIPEP,
                  prottabledata.HEADER_NO_PEPTIDE,
                  prottabledata.HEADER_NO_PSM,
                  ]
    for field in basefields:
        allfields[field] = False
    for field in poolfields:
        allfields[field] = poolnames
    return allfields


def get_bestpeptide_fields(poolnames=False):
    return {prottabledata.HEADER_QSCORE: poolnames}



#!/usr/bin/python3

import sys
from app.drivers import startup
from app.drivers.prottable import (probability, info, merge, isoquant, fdr,
                                   precursorarea, create_empty, bestpeptide)


def main():
    drivers = [info.AddProteinInfoDriver(),
               merge.BuildProteinTableDriver(),
               isoquant.AddIsobaricQuantDriver(),
               precursorarea.AddPrecursorAreaDriver(),
               probability.AddProteinProbability(),
               create_empty.CreateEmptyDriver(),
               bestpeptide.BestPeptidePerProtein(),
               fdr.ProttableFDRDriver(),
               fdr.PickedFDRDriver(),
               fdr.ProttableAddFDRDriver(),
               ]
    startup.start_msstitch(drivers, sys.argv)



from app.drivers.mslookup import base
from app.readers import tsv as tsvreader
from app.actions.mslookup import psms as lookup
from app.drivers.options import mslookup_options


class PSMLookupDriver(base.LookupDriver):
    lookuptype = 'psm'
    command = 'psms'
    commandhelp = ('Loads PSM table into lookup. Important for '
                   'several steps later on, such as protein grouping and '
                   'PSM quantitation. PSM TSV table passed to -i, '
                   'With flag --unroll, --spectracol, and specify a FASTA '
                   'file with --fasta, a Biomart map file with --map. When '
                   'using --map for a decoy lookup, use --decoy.')

    def __init__(self):
        super().__init__()
        self.infiletype = 'TSV PSM table (MSGF+)'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['spectracol', 'mapfn',
                                                 'decoy', 'fasta', 'unroll',
                                                 'genefield', 'fastadelim'],
                                                mslookup_options))

    def create_lookup(self):
        header = tsvreader.get_tsv_header(self.fn)
        specfncol = header[int(self.spectracol) - 1]
        fastadelim, genefield = self.get_fastadelim_genefield(self.fastadelim,
                                                              self.genefield)
        lookup.create_psm_lookup(self.fn, self.fasta, self.mapfn, header,
                                 self.lookup, self.unroll, specfncol,
                                 self.decoy, fastadelim, genefield)



from app.drivers import base
from app.readers import pycolator as readers
from app.readers import xml
from app.writers import pycolator as writers


class PycolatorDriver(base.BaseDriver):
    """Driver for pycolator functions"""
    def __init__(self):
        super().__init__()
        self.infiletype = 'percolator out XML'

    def prepare_percolator_output(self, fn):
        """Returns namespace and static xml from percolator output file"""
        ns = xml.get_namespace(fn)
        static = readers.get_percolator_static_xml(fn, ns)
        return ns, static

    def get_all_peptides(self):
        return readers.generate_peptides(self.fn, self.ns)

    def get_all_psms(self):
        return readers.generate_psms(self.fn, self.ns)

    def get_all_psms_strings(self):
        return readers.generate_psms_multiple_fractions_strings([self.fn],
                                                                self.ns)

    def get_all_peptides_strings(self):
        return readers.generate_peptides_multiple_fractions_strings([self.fn],
                                                                    self.ns)

    def prepare(self):
        self.ns, self.static_xml = self.prepare_percolator_output(self.fn)
        self.allpeps = self.get_all_peptides()
        self.allpsms = self.get_all_psms()

    def run(self):
        self.prepare()
        self.set_features()
        self.write()
        self.finish()

    def write(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        writers.write_percolator_xml(self.static_xml, self.features, outfn)



import os
from lxml import etree
from statistics import median

from app.dataformats import mzidtsv as constants
from tests.integration import basetests


class TestAddSpecData(basetests.MzidTSVBaseTest):
    command = 'specdata'
    suffix = '_spectradata.tsv'
    infilename = 'mzidtsv_filtered_fr1-2_nospecdata.txt'

    def test_addspecdata(self):
        options = ['--dbfile', self.dbfile, '--spectracol', '2']
        self.run_command(options)
        sql = ('SELECT pr.rownr, bs.set_name, sp.retention_time, '
               'sp.ion_injection_time '
               'FROM psmrows AS pr JOIN psms USING(psm_id) '
               'JOIN mzml AS sp USING(spectra_id) '
               'JOIN mzmlfiles USING(mzmlfile_id) '
               'JOIN biosets AS bs USING(set_id) '
               'ORDER BY pr.rownr')
        fields = ['Biological set', 'Retention time(min)',
                  'Ion injection time(ms)']
        expected_values = self.process_dbvalues_both(self.dbfile, sql, [],
                                                     [1, 2, 3], fields)
        self.check_results(fields, self.rowify(expected_values))


class TestQuantTSV(basetests.MzidTSVBaseTest):
    command = 'quant'
    suffix = '_quant.tsv'

    def test_quanttsv_isobaric(self):
        options = ['--dbfile', self.dbfile, '--isobaric']
        self.run_command(options)
        sql = ('SELECT pr.rownr, ic.channel_name, iq.intensity '
               'FROM psmrows AS pr JOIN psms USING(psm_id) '
               'JOIN isobaric_quant AS iq USING(spectra_id) '
               'JOIN isobaric_channels AS ic USING(channel_id)')
        expected_values = self.get_values_from_db(self.dbfile, sql)
        fields = ['tmt10plex_{}'.format(ch) for ch in ['126', '127N', '127C',
                                                       '128N', '128C', '129N',
                                                       '129C', '130N', '130C',
                                                       '131']]
        self.check_results(fields, self.rowify(expected_values))

    def test_quanttsv_precursor(self):
        dbfile = os.path.join(self.fixdir, 'mzidtsv_db.sqlite')
        options = ['--dbfile', dbfile, '--precursor']
        self.run_command(options)
        sql = ('SELECT pr.rownr, pq.intensity '
               'FROM psmrows AS pr JOIN psms USING(psm_id) '
               'LEFT OUTER JOIN ms1_align USING(spectra_id) '
               'LEFT OUTER JOIN ms1_quant AS pq USING(feature_id)')
        expected_values = self.get_values_from_db(self.dbfile, sql)
        self.check_results(['MS1 area'], self.rowify(expected_values))

    def test_quanttsv_both(self):
        dbfile = os.path.join(self.fixdir, 'mzidtsv_db.sqlite')
        options = ['--dbfile', dbfile, '--isobaric', '--precursor']
        self.run_command(options)
        sql = ('SELECT pr.rownr, ic.channel_name, iq.intensity, pq.intensity '
               'FROM psmrows AS pr JOIN psms USING(psm_id) '
               'JOIN isobaric_quant AS iq USING(spectra_id) '
               'JOIN isobaric_channels AS ic USING(channel_id) '
               'LEFT OUTER JOIN ms1_align USING(spectra_id) '
               'LEFT OUTER JOIN ms1_quant AS pq USING(feature_id) '
               'ORDER BY pr.rownr')
        expected_values = self.process_dbvalues_both(self.dbfile, sql, [1, 2],
                                                     [3], ['MS1 area'])
        fields = ['tmt10plex_{}'.format(ch) for ch in ['126', '127N', '127C',
                                                       '128N', '128C', '129N',
                                                       '129C', '130N', '130C',
                                                       '131']]
        fields.append('MS1 area')
        self.check_results(fields, self.rowify(expected_values))


class TestPercoTSV(basetests.MzidTSVBaseTest):
    command = 'percolator'
    suffix = '_percolated.tsv'
    infilename = 'mzidtsv_fr0.txt'
    field_p_map = {'percolator svm-score': 'score',
                   'PSM p-value': 'psm_p_value',
                   'PSM q-value': 'psm_q_value',
                   'PSM-PEP': 'psm_pep',
                   'peptide q-value': 'peptide_q_value',
                   'peptide PEP': 'peptide_pep',
                   }

    def test_add_percolator(self):
        mzidfn = os.path.join(self.fixdir, 'msgfperco_fr0.mzid')
        options = ['--mzid', mzidfn]
        self.run_command(options)
        expected = self.get_percolator_from_msgf(mzidfn,
                                                 self.field_p_map.keys())
        self.check_results(self.field_p_map.keys(), expected)

    def get_percolator_from_msgf(self, msgffile, checkfields):
        count = 0
        ns = self.get_xml_namespace(msgffile)
        for ac, specidres in etree.iterparse(
                msgffile, tag='{%s}'
                'SpectrumIdentificationResult' % ns['xmlns']):
            for result in specidres.findall(
                    '{%s}SpectrumIdentificationItem' % ns['xmlns']):
                perco = [x for x in
                         result.findall('{%s}userParam' % ns['xmlns'])
                         if x.attrib['name'].split(':')[0] == 'percolator']
                perco = {x.attrib['name'].replace('percolator:', ''):
                         x.attrib['value'] for x in perco}
                if not perco:
                    perco = {key: None for key in self.field_p_map.values()}
                #outresult = [(count, 'SpecID', sid)]
                outresult = []
                outresult.extend([(count, field,
                                   perco[self.field_p_map[field]])
                                  for field in checkfields])
                yield outresult
                count += 1


class TestMergeTSV(basetests.MzidTSVBaseTest):
    command = 'merge'
    suffix = '_concat.tsv'
    infilename = 'mzidtsv_fr0.txt'

    def test_mergetsv(self):
        self.infile = [self.infile, os.path.join(self.fixdir,
                                                 'mzidtsv_fr1.txt')]
        self.run_command()
        resultlines = self.get_all_lines(self.resultfn)
        for expectfn in self.infile:
            for line in self.get_all_lines(expectfn):
                self.assertEqual(line, next(resultlines))


class TestSplitTSV(basetests.MzidTSVBaseTest):
    infilename = 'mzidtsv_filtered_fr1-2.txt'
    command = 'split'
    suffix = '_split.tsv'

    def setUp(self):
        super().setUp()
        self.resultfn = None
        self.expectlines = self.get_all_lines(self.infile)

    def test_auto_bioset_column(self):
        self.run_command(['--bioset'])
        resultfn = os.path.join(self.workdir, 'S1.tsv')
        for line in self.get_all_lines(resultfn):
            self.assertEqual(line, next(self.expectlines))

    def test_splitcol(self):
        setnames = ['dataset_17694.dat_task_0.mzml',
                    'dataset_17694.dat_task_1.mzml']
        options = ['--splitcol', '1']
        self.run_command(options)
        resultfiles = [os.path.join(self.workdir, '{}.tsv'.format(setname))
                       for setname in setnames]
        for resultfn in resultfiles:
            for line in self.get_all_lines(resultfn):
                self.assertEqual(line, next(self.expectlines))


class TestConffiltTSV(basetests.MzidTSVBaseTest):
    command = 'conffilt'
    infilename = 'mzidtsv_fr0.txt'
    suffix = '_filtconf.txt'

    def test_confidence_filter_lower(self):
        conflvl = 0
        self.run_conffilt(conflvl, 'lower', confcol=14)

    def test_confidence_filter_lower_confpattern(self):
        conflvl = 0
        self.run_conffilt(conflvl, 'lower', confpat='EValue')

    def test_confidence_filter_higher(self):
        conflvl = 0
        self.run_conffilt(conflvl, 'higher', confcol=14)

    def run_conffilt(self, conflvl, better, confcol=False, confpat=False):
        options = ['--confidence-better', better,
                   '--confidence-lvl', str(conflvl)]
        if confcol is not False:
            options.extend(['--confidence-col', str(confcol)])
        elif confpat:
            options.extend(['--confcolpattern', confpat])
        print(options)
        self.run_command(options)
        asserter = {'lower': self.assertLess,
                    'higher': self.assertGreater}[better]
        for line in self.get_all_lines(self.resultfn):
            asserter(float(line.strip('\n').split('\t')[confcol - 1]), conflvl)

    def test_confidence_omit_confcol(self):
        options = ['--confidence-better', 'lower', '--confidence-lvl', '0.01']
        self.run_command_expect_error(options)

    def test_omit_conf_better(self):
        options = ['--confidence-col', '18', '--confidence-lvl', '0.01']
        self.run_command_expect_error(options)

    def test_omit_conf_val(self):
        options = ['--confidence-col', '18', '--confidence-better', 'lower']
        self.run_command_expect_error(options)


class TestProteinGroup(basetests.MzidTSVBaseTest):
    command = 'proteingroup'
    infilename = 'mzidtsv_filtered_fr1-2.txt'
    suffix = '_protgroups.txt'
    dbfile = 'mzidtsv_db.sqlite'

    def run_and_analyze(self, options):
        self.run_command(options)
        result = self.parse_proteingroups(self.resultfn)
        expected = self.parse_proteingroups(
            os.path.join(self.fixdir,
                         'mzidtsv_filtered_fr1-2_proteingrouped.txt'))
        self.do_asserting(result, expected)

    def parse_proteingroups(self, fn):
        with open(fn) as fp:
            header = next(fp).strip().split('\t')
            master_ix = header.index(constants.HEADER_MASTER_PROT)
            pgcontent_ix = header.index(constants.HEADER_PG_CONTENT)
            pgamount_ix = header.index(constants.HEADER_PG_AMOUNT_PROTEIN_HITS)
            for line in fp:
                line = line.strip().split('\t')
                yield {'master': line[master_ix],
                       'content': line[pgcontent_ix],
                       'amount': line[pgamount_ix],
                       }

    def do_asserting(self, result, expected, unrolled=False):
        for res, exp in zip(result, expected):
            self.assertEqual(set(res['master'].split(';')),
                             set(exp['master'].split(';')))
            self.assertEqual(res['amount'], exp['amount'])
            rescontent = res['content'].split(';')
            expcontent = exp['content'].split(';')
            self.assertEqual(set(rescontent), set(expcontent))

    def test_proteingroups(self):
        options = ['--dbfile', self.dbfile]
        self.expected = None
        self.run_and_analyze(options)


class TestAddGenes(basetests.MzidTSVBaseTest):
    command = 'genes'
    suffix = '_genes.txt'
    infilename = 'mzidtsv_filtered_fr1-2.txt'

    def test_addgenes(self):
        self.run_command(['--dbfile', self.dbfile])
        for line in self.get_values(['Gene', 'Gene Symbol', 'Description',
                                     'Protein']):
            genes = line[0][2].split(';')
            assoc_ids = line[1][2].split(';')
            descriptions = ['{}]'.format(x).replace(']]', ']')
                            for x in line[2][2].split('];')]
            proteins = [x.split('(')[0] for x in line[3][2].split(';')]
            sql = ('SELECT p.protein_acc, g.gene_acc, a.assoc_id, '
                   'd.description FROM proteins AS p '
                   'JOIN genes AS g USING(protein_acc) '
                   'JOIN associated_ids AS a ON p.protein_acc=a.protein_acc'
                   ' JOIN prot_desc AS d ON d.protein_acc=p.protein_acc '
                   'WHERE p.protein_acc IN ({})')
            dbvals = self.get_values_from_db(self.dbfile, sql.format(
                ','.join(['"{}"'.format(x) for x in proteins])))
            exp_g, exp_assoc, exp_desc = set(), set(), set()
            for prot, gene, aid, desc in dbvals:
                exp_g.add(gene)
                exp_assoc.add(aid)
                exp_desc.add(desc)
            for exp_set, result in zip([exp_g, exp_assoc, exp_desc],
                                       [genes, assoc_ids, descriptions]):
                self.assertEqual(0, len(exp_set.difference(result)))


class TestIso(basetests.MzidTSVBaseTest):

    def get_denominator(self, line, denom_ch):
        denomvals = [float(line[ch]) for ch in denom_ch if line[ch] != 'NA']
        return sum(denomvals) / len(denomvals)

    def get_infile_lines(self, infile=None):
        if infile is None:
            infile = self.infile[0]
        with open(infile) as fp:
            header = next(fp).strip('\n').split('\t')
            for line in fp:
                line = line.strip('\n').split('\t')
                yield {field: val for field, val in zip(header, line)}

    def check_normalize_medians(self, channels, denom_ch, minint, stdout,
                                medianpsms):
        ch_medians = {ch: [] for ch in channels}
        for line in self.get_infile_lines(medianpsms):
            line.update({ch: line[ch]
                         if line[ch] != 'NA' and float(line[ch]) > minint
                         else 'NA' for ch in channels})
            denom = self.get_denominator(line, denom_ch)
            if denom == 0:
                continue
            for ch in channels:
                if line[ch] == 'NA':
                    continue
                ch_medians[ch].append(float(line[ch]) / denom)
        ch_medians = {ch: median(vals) for ch, vals in ch_medians.items()}
        stdout = stdout.decode().split('\n')
        self.assertEqual(stdout[0],
                         'Channel intensity medians used for normalization:')
        stdout_channels = {x.split(' - ')[0]: x.split(' - ')[1]
                           for x in stdout[1:]}
        for ch in channels:
            self.assertEqual(float(stdout_channels[ch]), ch_medians[ch])
        return ch_medians

    def do_check(self, minint, stdout, normalize=False, medianpsms=None,
                 resultch=False):
        channels = ['fake_ch{}'.format(x) for x in range(8)]
        # TODO only for backwards compatibilty, remove if statement around
        # assignment when msspsmtable isonormalize is removed
        if not resultch:
            resultch = ['ratio_{}'.format(x) for x in channels]
        denom_ch = channels[0:2]
        if normalize:
            ch_medians = self.check_normalize_medians(channels, denom_ch,
                                                      minint, stdout,
                                                      medianpsms)
        for in_line, resultline in zip(self.get_infile_lines(),
                                       self.get_values(resultch)):
            in_line.update({ch: in_line[ch]
                            if in_line[ch] != 'NA' and
                            float(in_line[ch]) > minint else 'NA'
                            for ch in channels})
            resultline = [x[2] for x in resultline]
            denom = self.get_denominator(in_line, denom_ch)
            if denom == 0:
                exp_line = ['NA'] * len(channels)
            elif normalize:
                exp_line = [str((float(in_line[ch]) / denom) / ch_medians[ch])
                            if in_line[ch] != 'NA' else 'NA'
                            for ch in channels]
            else:
                exp_line = [str((float(in_line[ch]) / denom))
                            if in_line[ch] != 'NA' else 'NA'
                            for ch in channels]
            self.assertEqual(resultline, exp_line)


class TestIsoRatio(TestIso):
    suffix = '_ratio_isobaric.txt'
    command = 'isoratio'
    infilename = 'mzidtsv.txt'

    def test_denomcolpattern(self):
        stdout = self.run_command_stdout(['--isobquantcolpattern', 'fake_ch',
                                          '--denompatterns', '_ch0', '_ch1'])
        self.do_check(0, stdout)

    def test_denomcolpattern_regex(self):
        stdout = self.run_command_stdout(['--isobquantcolpattern', 'fake_ch',
                                          '--denompatterns', '_ch[0-1]'])
        self.do_check(0, stdout)


class TestIsoFeatRatio(TestIso):
    suffix = '_ratio_isobaric.txt'
    command = 'isoratio'
    infilename = 'mzidtsv_intensities.txt'
    channels = ['tmt10plex_{}'.format(x) for x in ['126', '127N', '127C',
                                                   '128N', '128C', '129N',
                                                   '129C', '130N', '130C',
                                                   '131']]
    nopsms = ['{} - # quanted PSMs'.format(ch) for ch in channels]

    def test_normalized_isoquant(self):
        options = ['--protcol', '14', '--isobquantcolpattern', 'tmt10plex',
                   '--denompatterns', '_126', '--normalize', 'median']
        self.run_command(options)
        self.isoquant_check(
            os.path.join(self.fixdir, 'prottable_normalized_isoquant.txt'),
            'Accession', self.channels, self.nopsms)

    def test_normalized_othertable_isoquant(self):
        prottable_ratiofn = os.path.join(self.fixdir,
                                         'mzidtsv_ratios.txt')
        options = ['--protcol', '14', '--isobquantcolpattern', 'tmt10plex',
                   '--denompatterns', '_126', '--normalize', 'median',
                   '--norm-ratios', prottable_ratiofn]
        self.run_command(options)
        self.isoquant_check(
            os.path.join(self.fixdir, 'prottable_normalized_isoquant.txt'),
            'Accession', self.channels, self.nopsms)

    def test_normalized_targettable_isoquant(self):
        prottable_targettable = os.path.join(self.fixdir,
                                             'prottable_only_acc.txt')
        options = ['--protcol', '14', '--isobquantcolpattern', 'tmt10plex',
                   '--denompatterns', '_126', '--normalize', 'median',
                   '--targettable', prottable_targettable]
        self.run_command(options)
        self.isoquant_check(
            os.path.join(self.fixdir, 'prottable_normalized_isoquant.txt'),
            'Protein accession', self.channels, self.nopsms)


class TestIsoNormalize(TestIso):
    suffix = '_normalized_isobaric.txt'
    command = 'isonormalize'
    infilename = 'mzidtsv.txt'
    channels = ['fake_ch{}'.format(x) for x in range(8)]

    def test_normalize(self):
        stdout = self.run_command_stdout(['--isobquantcolpattern', 'fake_ch',
                                          '--denomcols', '21', '22'])
        self.do_check(0, stdout, normalize=True, resultch=self.channels)

    def test_normalize_minint(self):
        minint = 3000
        stdout = self.run_command_stdout(['--isobquantcolpattern', 'fake_ch',
                                          '--denomcols', '21', '22',
                                          '--minint', str(minint)])
        self.do_check(minint, stdout, normalize=True, resultch=self.channels)


class TestIsoNormalizeTwofiles(TestIso):
    infilename = 'mzidtsv_short.txt'
    suffix = '_normalized_isobaric.txt'
    command = 'isonormalize'
    channels = ['fake_ch{}'.format(x) for x in range(8)]

    def test_two_psm_files(self):
        """Tests calculating medians on different file than the one doing the
        median centering on"""
        medianpsms = os.path.join(self.fixdir, 'mzidtsv.txt')
        stdout = self.run_command_stdout(['--isobquantcolpattern', 'fake_ch',
                                          '--denomcols', '21', '22',
                                          '--medianpsms', medianpsms])
        self.do_check(0, stdout, normalize=True, medianpsms=medianpsms,
                      resultch=self.channels)



from app.readers import xmlformatting as formatting
from app.readers import pycolator as readers


def get_score(elements, ns, scoretype='svm_score'):
    for el in elements:
        score = el.xpath('xmlns:{0}'.format(scoretype), namespaces=ns)[0].text
        formatting.clear_el(el)
        yield score


def prepare_qvality_input(fn, feattype, get_static_percolator_output):
    featextractors = {'peptide': readers.generate_peptides,
                      'psm': readers.generate_psms
                      }

    ns, static_xml = get_static_percolator_output(fn)
    features_for_qvality = featextractors[feattype](fn, ns)
    return get_score(features_for_qvality, ns)



from app.drivers.mzidtsv import MzidTSVDriver
from app.drivers.options import mzidtsv_options
from app.actions.mzidtsv import isonormalize as prep
from app.dataformats import prottable as prottabledata
from app.readers import tsv


class PSMIsoquantRatioDriver(MzidTSVDriver):
    outsuffix = '_ratio_isobaric.txt'
    command = 'isoratio'
    commandhelp = ('Produce isobaric PSM ratios from a '
                   'table containing raw intensities.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['quantcolpattern',
                                                 'denompatterns', 'denomcols',
                                                 'minint', 'targettable',
                                                 'proteincol', 'normalize',
                                                 'normalizeratios'],
                                                mzidtsv_options))

    def get_psms(self):
        if self.denomcols is not None:
            denomcols = [self.number_to_headerfield(col, self.oldheader)
                         for col in self.denomcols]
        elif self.denompatterns is not None:
            denomcolnrs = [tsv.get_columns_by_pattern(self.oldheader, pattern)
                           for pattern in self.denompatterns]
            denomcols = set([col for cols in denomcolnrs for col in cols])
        else:
            raise RuntimeError('Must define either denominator column numbers '
                               'or regex pattterns to find them')
        quantcols = tsv.get_columns_by_pattern(self.oldheader,
                                               self.quantcolpattern)
        self.get_column_header_for_number(['proteincol'], self.oldheader)
        nopsms = [prep.get_no_psms_field(qf) for qf in quantcols]
        if self.proteincol and self.targettable:
            targetheader = tsv.get_tsv_header(self.targettable)
            self.header = targetheader + quantcols + nopsms
        elif not self.proteincol and not self.targettable:
            self.header = (self.oldheader +
                           ['ratio_{}'.format(x) for x in quantcols])
        elif self.proteincol and not self.targettable:
            self.header = [prottabledata.HEADER_ACCESSION] + quantcols + nopsms
        self.psms = prep.get_isobaric_ratios(self.fn, self.oldheader,
                                             quantcols, denomcols, self.minint,
                                             self.targettable, self.proteincol,
                                             self.normalize,
                                             self.normalizeratios)


class PSMIsoquantNormalizeDriver(MzidTSVDriver):
    outsuffix = '_normalized_isobaric.txt'
    command = 'isonormalize'
    commandhelp = ('Produce median normalized isobaric PSM ratios from a '
                   'table containing raw intensities.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['quantcolpattern',
                                                 'medianpsms', 'denompatterns',
                                                 'denomcols', 'minint'],
                                                mzidtsv_options))

    def get_psms(self):
        self.header = self.oldheader[:]
        if self.denomcols is not None:
            denomcols = [self.number_to_headerfield(col, self.oldheader)
                         for col in self.denomcols]
        elif self.denompatterns is not None:
            denomcolnrs = [tsv.get_columns_by_pattern(self.oldheader, pattern)
                           for pattern in self.denompatterns]
            denomcols = set([col for cols in denomcolnrs for col in cols])
        else:
            raise RuntimeError('Must define either denominator column numbers '
                               'or regex pattterns to find them')
        quantcols = tsv.get_columns_by_pattern(self.oldheader,
                                               self.quantcolpattern)
        if self.medianpsms is not None:
            medianheader = tsv.get_tsv_header(self.medianpsms)
        else:
            medianheader = False
        self.psms = prep.get_normalized_ratios(self.fn, self.oldheader,
                                               quantcols, denomcols,
                                               self.minint, self.medianpsms,
                                               medianheader)



from tests.integration import basetests

import os
import sqlite3
import yaml
from Bio import SeqIO
from lxml import etree


class SearchspaceLookup(basetests.BaseTest):
    suffix = ''
    infilename = 'proteins.fasta'
    executable = 'msslookup'


class TestTrypticLookup(SearchspaceLookup):
    command = 'seqspace'

    def all_seqs_in_db(self, dbfn, sequences, seqtype):
        db = sqlite3.connect(dbfn)
        seqs_in_db = set()
        for seq in sequences:
            seqs_in_db.add(self.seq_in_db(db, seq, seqtype))
        db.close()
        return seqs_in_db == set([True])

    def query_db_assert(self, options=None, seqtype=None):
        if options is None:
            options = []
        with open(os.path.join(self.fixdir, 'peptides_trypsinized.yml')) as fp:
            tryp_sequences = yaml.load(fp)
        sequences = tryp_sequences['fully_tryptic']
        if seqtype is not None:
            sequences.extend(tryp_sequences[seqtype])
        self.run_command(options)
        self.assertTrue(self.all_seqs_in_db(self.resultfn,
                                            sequences, seqtype))

    def run_without_db(self, options=None, seqtype=None):
        self.resultfn = os.path.join(self.workdir,
                                     'mslookup_db.sqlite')
        self.query_db_assert(options, seqtype)

    def run_with_existing_db(self, options=None, seqtype=None):
        if options is None:
            options = []
        self.resultfn = os.path.join(self.workdir, 'seqspace.db')
        options.extend(['--dbfile', self.resultfn])
        self.copy_db_to_workdir('mzidtsv_db.sqlite')
        self.query_db_assert(options, seqtype)

    def test_cutproline_nodb(self):
        self.run_without_db(['--cutproline'], 'proline_cuts')

    def test_cutproline_yesdb(self):
        self.run_with_existing_db(['--cutproline'], 'proline_cuts')

    def test_ntermwildcards_nodb(self):
        self.run_without_db(['--insourcefrag'], 'ntermfalloff')

    def test_ntermwildcards_yes_db(self):
        self.run_with_existing_db(['--insourcefrag'], 'ntermfalloff')

    def test_noflags_no_db(self):
        self.run_without_db()

    def test_noflags_yes_db(self):
        self.run_with_existing_db()


class TestWholeProteinSeqLookup(SearchspaceLookup):
    command = 'protspace'

    def all_seqs_in_db(self, dbfn, sequences, seqtype):
        db = sqlite3.connect(dbfn)
        seqs_in_db = set()
        sql = ('SELECT EXISTS(SELECT seq FROM protein_peptides WHERE '
               'seq=? LIMIT 1)')
        for seq in sequences:
            seqs_in_db.add(db.execute(sql, (seq,)).fetchone()[0] == 1)
        db.close()
        return seqs_in_db == set([True])

    def query_db_assert(self, options):
        with open(os.path.join(self.fixdir, 'allpeptides_proteins.yml')) as fp:
            sequences = [x.replace('L', 'I') for x in yaml.load(fp)]
        options.extend(['--minlen', '6'])
        self.run_command(options)
        self.assertTrue(self.all_seqs_in_db(self.resultfn, sequences,
                                            seqtype=False))

    def test_without_db(self, seqtype=None):
        self.resultfn = os.path.join(self.workdir,
                                     'mslookup_db.sqlite')
        self.query_db_assert([])

    def test_with_existing_db(self, seqtype=None):
        self.resultfn = os.path.join(self.workdir, 'seqspace.db')
        options = ['--dbfile', self.resultfn]
        self.copy_db_to_workdir('mzidtsv_db.sqlite')
        self.query_db_assert(options)


class TestSpectraLookup(basetests.MSLookupTest):
    command = 'spectra'
    infilename = 'few_spectra.mzML'

    def check_spectra(self, bsets):
        sql = ('SELECT mf.mzmlfilename, bs.set_name, s.scan_nr, s.charge, '
               's.mz, s.retention_time, s.ion_injection_time, s.spectra_id '
               'FROM mzml AS s '
               'JOIN mzmlfiles AS mf USING(mzmlfile_id) '
               'JOIN biosets AS bs USING(set_id)')
        specrecs = {}
        for rec in self.get_values_from_db(self.resultfn, sql):
            specrecs[rec[2]] = {'fn': rec[0], 'bs': rec[1], 'charge': rec[3],
                                'mz': rec[4], 'rt': rec[5], 'iit': rec[6],
                                'sid': rec[7]}
        for scannr, spec in self.get_spectra_mzml(self.infile, bsets):
            self.assertEqual(spec, specrecs[scannr])

    def get_spectra_mzml(self, infiles, bsets):
        def get_cvparam_value(parent, name, ns):
            return [x.attrib['value'] for x in
                    parent.findall('{%s}cvParam' % ns['xmlns'])
                    if x.attrib['name'] == name]

        def multifind(elements, parent, ns):
            element = elements.pop(-1)
            if len(elements) > 0:
                parent = multifind(elements, parent, ns)
            return parent.find('{%s}%s' % (ns['xmlns'], element))

        for mzfnid, (bset, infile) in enumerate(zip(bsets, infiles)):
            ns = self.get_xml_namespace(infile)
            for ac, spectrum in etree.iterparse(
                    infile, tag='{%s}spectrum' % ns['xmlns']):
                if get_cvparam_value(spectrum, 'ms level', ns)[0] != "2":
                    continue
                scannr = spectrum.attrib['id'].split()[2].split('=')[1]
                scan = multifind(['scanList', 'scan'], spectrum, ns)
                precursor = multifind(['precursorList', 'precursor',
                                       'selectedIonList', 'selectedIon'],
                                      spectrum, ns)
                rt = get_cvparam_value(scan, 'scan start time', ns)[0]
                iit = get_cvparam_value(scan, 'ion injection time', ns)[0]
                charge = get_cvparam_value(precursor, 'charge state', ns)[0]
                mz = get_cvparam_value(precursor, 'selected ion m/z', ns)[0]
                exp_data = {'fn': os.path.basename(infile), 'bs': bset,
                            'charge': int(charge),
                            'sid': '{}_{}'.format(mzfnid + 1, scannr),
                            'mz': float(mz), 'rt': float(rt), 'iit': float(iit)}
                yield scannr, exp_data

    def test_spectra_newdb(self):
        setnames = ['Set1']
        options = ['--setnames']
        options.extend(setnames)
        self.run_command(options)
        self.check_spectra(setnames)

    def test_spectra_olddb(self):
        self.copy_db_to_workdir('bioset_lookup.sqlite')
        setnames = ['Set1']
        options = ['--setnames']
        options.extend(setnames)
        self.run_command(options)
        self.check_spectra(setnames)


class TestPSMLookup(basetests.MSLookupTest):
    command = 'psms'
    base_db_fn = 'spectra_lookup.sqlite'

    def check_db_fasta(self, fasta, exp_proteins=None, desc=True):
        if exp_proteins is None:
            exp_proteins = {rec.id: {'seq': rec.seq, 'desc': rec.description}
                            for rec in SeqIO.parse(fasta, 'fasta')}
        self.check_db_base(exp_proteins)
        sql = ('SELECT ps.protein_acc, ps.sequence, pd.description '
               'FROM protein_seq AS ps '
               'JOIN prot_desc AS pd USING(protein_acc)')
        if not desc:
            sql = ('SELECT ps.protein_acc, ps.sequence '
                   'FROM protein_seq AS ps '
                   'JOIN prot_desc AS pd USING(protein_acc)')
        for result_protein in self.get_values_from_db(self.resultfn, sql):
            self.assertEqual(exp_proteins[result_protein[0]]['seq'],
                             result_protein[1])
            if desc:
                self.assertEqual(exp_proteins[result_protein[0]]['desc'],
                                 result_protein[2])

    def check_db_base(self, expected_proteins=None):
        expected_psms = self.get_expected_psms()
        if expected_proteins is None:
            expected_proteins = {x for y in expected_psms.values()
                                 for x in y['proteins']}
        protsql = 'SELECT * FROM proteins'
        for protein in self.get_values_from_db(self.resultfn, protsql):
            self.assertIn(protein[1], expected_proteins)
        psmsql = ('SELECT ps.sequence, p.score, pr.rownr '
                  'FROM psmrows AS pr JOIN psms AS p USING(psm_id) '
                  'JOIN peptide_sequences AS ps USING(pep_id)')
        for psm in self.get_values_from_db(self.resultfn, psmsql):
            expected_psm = (expected_psms[psm[2]]['seq'],
                            expected_psms[psm[2]]['score'])
            self.assertEqual((psm[0], psm[1]), expected_psm)
        ppsql = ('SELECT pp.protein_acc, pr.rownr FROM psmrows AS pr '
                 'JOIN protein_psm AS pp USING(psm_id)')
        for protpsm in self.get_values_from_db(self.resultfn, ppsql):
            self.assertIn(protpsm[0], expected_psms[protpsm[1]]['proteins'])

    def get_expected_psms(self):
        header = self.get_tsvheader(self.infile[0])
        prot_ix = header.index('Protein')
        seq_ix = header.index('Peptide')
        score_ix = header.index('MSGFScore')
        psms = {}
        for row, line in enumerate(self.get_all_lines(self.infile[0])):
            line = line.strip('\n').split('\t')
            psms[row] = {'proteins': [x.split('(pre')[0] for x in
                                      line[prot_ix].split(';')],
                         'seq': line[seq_ix],
                         'score': line[score_ix],
                         }
        return psms


class TestPSMLookupEnsemblBase(TestPSMLookup):
    def check_db_map(self, fasta, martmap, versioned=False):
        exp_proteins = {rec.id: {'seq': rec.seq}
                        for rec in SeqIO.parse(fasta, 'fasta')}
        header = self.get_tsvheader(martmap)
        try:
            pix = header.index('Ensembl Protein ID')
        except ValueError:
            # new biomart version (2017 jan) has no "Ensembl" in header fields
            pix = header.index('Protein ID')
            gix = header.index('Gene ID')
        else:
            gix = header.index('Ensembl Gene ID')
        six = header.index('HGNC symbol')
        dix = header.index('Description')
        for line in self.get_all_lines(martmap):
            line = line.strip('\n').split('\t')
            if versioned:
                pacc = '{}.2'.format(line[pix])
            else:
                pacc = line[pix]
            exp_proteins[pacc].update({'gene': line[gix],
                                       'symb': line[six],
                                       'desc': line[dix],
                                       })
        self.check_db_fasta(fasta, exp_proteins)
        sql = ('SELECT g.protein_acc, g.gene_acc, aid.assoc_id, pd.description'
               ' FROM genes AS g JOIN associated_ids AS aid USING(protein_acc)'
               ' JOIN prot_desc AS pd USING(protein_acc)')
        count = 0
        for prot, gene, aid, desc in self.get_values_from_db(self.resultfn,
                                                             sql):
            count += 1
            self.assertEqual(gene, exp_proteins[prot]['gene'])
            self.assertEqual(aid, exp_proteins[prot]['symb'])
            self.assertEqual(desc, exp_proteins[prot]['desc'])
        self.assertNotEqual(count, 0)

    def check_db_decoy(self, fasta, martmap):
        def decoymod(accession):
            return 'decoy_{}'.format(accession)
        exp_proteins = {rec.id: {'seq': rec.seq}
                        for rec in SeqIO.parse(fasta, 'fasta')}
        header = self.get_tsvheader(martmap)
        pix = header.index('Ensembl Protein ID')
        gix = header.index('Ensembl Gene ID')
        six = header.index('HGNC symbol')
        dix = header.index('Description')
        for line in self.get_all_lines(martmap):
            line = line.strip('\n').split('\t')
            prot = decoymod(line[pix])
            exp_proteins[prot].update({'gene': decoymod(line[gix]),
                                       'symb': decoymod(line[six]),
                                       'desc': decoymod(line[dix]),
                                       })
        self.check_db_fasta(fasta, exp_proteins, desc=False)
        sql = ('SELECT g.protein_acc, g.gene_acc, aid.assoc_id '
               'FROM genes AS g JOIN associated_ids AS aid USING(protein_acc)')
        for prot, gene, aid in self.get_values_from_db(self.resultfn, sql):
            self.assertEqual(gene, exp_proteins[prot]['gene'])
            self.assertEqual(aid, exp_proteins[prot]['symb'])


class TestPSMLookupEnsembl(TestPSMLookupEnsemblBase):
    infilename = 'mzidtsv_filtered_fr1-2.txt'

    def test_no_fasta(self):
        options = ['--spectracol', '2']
        self.run_command(options)
        self.check_db_base()

    def test_with_fasta(self):
        fastafn = os.path.join(self.fixdir, 'ensembl.fasta')
        options = ['--spectracol', '2', '--fasta', fastafn]
        self.run_command(options)
        self.check_db_fasta(fastafn)

    def test_fasta_map(self):
        fastafn = os.path.join(self.fixdir, 'ensembl.fasta')
        mapfn = os.path.join(self.fixdir, 'biomart.map')
        options = ['--spectracol', '2', '--fasta', fastafn, '--map', mapfn]
        self.run_command(options)
        self.check_db_map(fastafn, mapfn)

    def test_newversion_fasta_map(self):
        fastafn = os.path.join(self.fixdir, 'ensembl.fasta')
        mapfn = os.path.join(self.fixdir, 'new_biomart.map')
        options = ['--spectracol', '2', '--fasta', fastafn, '--map', mapfn]
        self.run_command(options)
        self.check_db_map(fastafn, mapfn)

    def test_fasta_map_decoy(self):
        self.infile = os.path.join(self.fixdir,
                                   'mzidtsv_filtered_fr1-2_decoy.txt')
        fastafn = os.path.join(self.fixdir, 'decoy_ensembl.fasta')
        mapfn = os.path.join(self.fixdir, 'biomart.map')
        options = ['--spectracol', '2', '--fasta', fastafn,
                   '--map', mapfn, '--decoy']
        self.run_command(options)
        self.check_db_decoy(fastafn, mapfn)


class TestPSMLookupEnsemblExtraPeptides(TestPSMLookupEnsemblBase):
    """For when a DB has been used to which a user has added e.g. novel
    peptides, which are not annotated to a protein"""

    infilename = 'mzidtsv_filtered_fr1-2_extrapep.txt'

    def test_no_fasta(self):
        options = ['--spectracol', '2']
        self.run_command(options)
        self.check_db_base()

    def test_with_fasta(self):
        fastafn = os.path.join(self.fixdir, 'ensembl_extrapep.fasta')
        options = ['--spectracol', '2', '--fasta', fastafn]
        self.run_command(options)
        self.check_db_fasta(fastafn)

    def test_fasta_map(self):
        fastafn = os.path.join(self.fixdir, 'ensembl_extrapep.fasta')
        mapfn = os.path.join(self.fixdir, 'biomart.map')
        options = ['--spectracol', '2', '--fasta', fastafn, '--map', mapfn]
        self.run_command(options)
        self.check_db_map(fastafn, mapfn)

    def test_fasta_map_decoy(self):
        self.infile = os.path.join(self.fixdir,
                                   'mzidtsv_filtered_fr1-2_extrapep_decoy.txt')
        fastafn = os.path.join(self.fixdir, 'decoy_ensembl_extrapep.fasta')
        mapfn = os.path.join(self.fixdir, 'biomart.map')
        options = ['--spectracol', '2', '--fasta', fastafn,
                   '--map', mapfn, '--decoy']
        self.run_command(options)
        self.check_db_decoy(fastafn, mapfn)


class TestPSMLookupEnsemblVersioned(TestPSMLookupEnsemblBase):
    infilename = 'mzidtsv_filtered_fr1-2_versioned_ensembl.txt'

    def test_fasta_map_versions(self):
        fastafn = os.path.join(self.fixdir, 'ensembl_versioned.fasta')
        mapfn = os.path.join(self.fixdir, 'biomart.map')
        options = ['--spectracol', '2', '--fasta', fastafn, '--map', mapfn]
        self.run_command(options)
        self.check_db_map(fastafn, mapfn, versioned=True)


class TestPSMLookupDelimitedFasta(TestPSMLookup):
    infilename = 'mzidtsv_filtered_fr1-2_notensembl.txt'

    def test_with_fasta_with_delim(self):
        fastafn = os.path.join(self.fixdir, 'delim_header.fasta')
        field, delim, delimname = '2', '|', 'pipe'
        options = ['--spectracol', '2', '--fasta', fastafn,
                   '--fastadelim', delimname, '--genefield', field]
        self.run_command(options)
        self.check_db_fasta_with_own_format_delimiter(fastafn, delim, field)

    def check_db_fasta_with_own_format_delimiter(self, fasta, delim, field):
        self.check_db_fasta(fasta)
        exp_proteins = {rec.id:
                        {'gene': rec.description.split(delim)[int(field) - 1]}
                        for rec in SeqIO.parse(fasta, 'fasta')}
        sql = ('SELECT g.protein_acc, g.gene_acc FROM genes AS g ')
        for prot, gene in self.get_values_from_db(self.resultfn, sql):
            self.assertEqual(gene, exp_proteins[prot]['gene'])


class TestIsoquantLookup(basetests.MSLookupTest):
    command = 'isoquant'
    infilename = 'tmt_quant.consxml'
    base_db_fn = 'spectra_lookup.sqlite'

    def get_quantch_map(self):
        qmap_xml = {}
        for ac, map_el in etree.iterparse(self.infile[0], tag='map'):
            channeldata = map_el.attrib
            qmap_xml[channeldata['id']] = channeldata['label']
        return qmap_xml

    def check_quantmap(self):
        qmap_xml = self.get_quantch_map()
        qmap_xml = [qmap_xml[x]for x in sorted(qmap_xml.keys())]
        sql = 'SELECT channel_name FROM isobaric_channels'
        for xml_channel, db_rec in zip(qmap_xml,
                                       self.get_values_from_db(self.resultfn,
                                                               sql)):
            self.assertEqual(xml_channel, db_rec[0])

    def check_quantification(self):
        sql = ('SELECT iq.intensity, ic.channel_name FROM isobaric_quant '
               'AS iq JOIN isobaric_channels AS ic USING(channel_id)')
        qch_map = self.get_quantch_map()
        dbtmt = self.get_values_from_db(self.resultfn, sql)
        for ac, xml_quant in etree.iterparse(self.infile[0],
                                             tag='consensusElement'):
            for element in xml_quant.findall('.//element'):
                qval, qchan = next(dbtmt)
                self.assertEqual(qch_map[element.attrib['map']], qchan)
                self.assertEqual(float(element.attrib['it']), qval)

    def test_isoquant(self):
        fakespectra = os.path.join(self.workdir, 'task_0_dataset_17694.dat')
        options = ['--spectra', fakespectra]
        self.run_command(options)
        self.check_quantmap()
        self.check_quantification()


class TestMS1KronikLookup(basetests.MSLookupTest):
    command = 'ms1quant'
    infilename = 'kronik.txt'
    base_db_fn = 'spectra_lookup.sqlite'

    def test_kronik(self):
        self.fakespfn = os.path.join(self.workdir, 'task_0_dataset_17694.dat')
        options = ['--quanttype', 'kronik', '--rttol', '5', '--mztol', '20',
                   '--mztoltype', 'ppm', '--spectra', self.fakespfn]
        self.run_command(options)
        self.check_feats_stored()

    def check_feats_stored(self):
        PROTON_MASS = 1.0072
        sql = ('SELECT * FROM ms1_quant')
        feats = {1: {}}
        for recid, fnid, rt, mz, charge, val in self.get_values_from_db(
                self.resultfn, sql):
            try:
                feats[fnid][rt][mz][charge] = val
            except KeyError:
                try:
                    feats[fnid][rt][mz] = {charge: val}
                except KeyError:
                    feats[fnid][rt] = {mz: {charge: val}}
        with open(self.infile[0]) as fp:
            next(fp).strip().split('\t')
            for line in fp:
                line = line.strip('\n').split('\t')
                charge = float(line[3])
                mz = (float(line[4]) + charge * PROTON_MASS) / charge
                resval = feats[1][float(line[10])][mz][charge]
                self.assertEqual(float(line[6]), resval)


class TestMS1OpenMSLookup(basetests.MSLookupTest):
    command = 'ms1quant'
    infilename = 'featfinder.featXML'
    base_db_fn = 'spectra_lookup.sqlite'

    def test_featfind(self):
        self.fail('Not implemented tests for storing featfinder data yet')


class TestProteingroupLookup(basetests.MSLookupTest):
    command = 'proteingroup'
    infilename = 'mzidtsv_filtered_fr1-2.txt'
    base_db_fn = 'mzidtsv_db_nopgroups.sqlite'

    def test_proteingroup(self):
        self.run_command()
        self.check_pgmasters()
        self.check_coverage()
        self.check_pg_content()
        self.check_psm_protein_group()

    def check(self, sql, keyfun, valfun):
        exp_file = os.path.join(self.fixdir, 'mzidtsv_db.sqlite')
        result = {keyfun(x): valfun(x) for x in
                  self.get_values_from_db(self.resultfn, sql)}
        expected = {keyfun(x): valfun(x) for x in
                    self.get_values_from_db(exp_file, sql)}
        for key, value in result.items():
            self.assertIn(key, expected.keys())
            self.assertEqual(value, expected[key])

    def check_coverage(self):
        sql = 'SELECT * FROM protein_coverage'
        self.check(sql, lambda x: x[0], lambda x: x[1])

    def check_psm_protein_group(self):
        sql = ('SELECT ppg.psm_id, pgm.protein_acc FROM psm_protein_groups '
               'AS ppg JOIN protein_group_master AS pgm USING(master_id)')
        self.check(sql, lambda x: x[0], lambda x: x[1])

    def check_pg_content(self):
        sql = ('SELECT pgm.protein_acc, pgc.protein_acc, pgc.peptide_count, '
               'pgc.psm_count, pgc.protein_score '
               'FROM protein_group_content AS pgc '
               'JOIN protein_group_master AS pgm USING(master_id) '
               'ORDER BY pgm.protein_acc, pgc.protein_acc')
        self.check(sql, lambda x: x[0], lambda x: x[1:])

    def check_pgmasters(self):
        sql = ('SELECT * FROM protein_group_master')
        self.check(sql, lambda x: x[1], lambda x: 1)


class TestProtPepTableLookup(basetests.MSLookupTest):
    def run_checks(self, options, sql, checkfeats):
        self.run_command(options)
        result = self.get_values_from_db(self.resultfn, sql)
        with open(self.infile[0]) as fp:
            header = next(fp).strip().split('\t')
            for res, exp in zip(result, fp):
                exp = {field: val for field, val in
                       zip(header, exp.strip('\n').split('\t'))}
                for i, key in enumerate(checkfeats):
                    self.check_float(key, exp, res[i])

    def check_float(self, key, exp, result):
        expected = self.get_float_or_na(exp[key])
        self.assertEqual(result, expected)

    def check_isobaric(self, sql):
        def do_check(result, expected, header):
            expected = {k: v for k, v in
                        zip(header, expected.strip('\n').split('\t'))}
            [self.assertIn(key, expected) for key in result.keys()]
            [self.check_float(key, expected, result[key]) for key in result]
        protein_acc, resultquant = None, {}
        with open(self.infile[0]) as fp:
            header = next(fp).strip().split('\t')
            for rec in self.get_values_from_db(self.resultfn, sql):
                if rec[0] != protein_acc and protein_acc is not None:
                    do_check(resultquant, next(fp), header)
                    resultquant = {}
                    protein_acc = rec[0]
                elif protein_acc is None:
                    protein_acc = rec[0]
                resultquant[rec[1]] = rec[2]


class TestPeptidetableLookup(TestProtPepTableLookup):
    command = 'peptides'
    infilename = 'peptable.txt'
    base_db_fn = 'mzidtsv_db.sqlite'

    def run_peptable(self, opts=None):
        options = ['--peptidecol', '13', '--setnames', 'S1',
                   '--isobquantcolpattern', 'tmt10plex', '--fdrcolpattern',
                   '^q-value', '--pepcolpattern', '^PEP', '--psmnrcolpattern',
                   'quanted', '--ms1quantcolpattern', 'area']
        if opts is not None:
            options.extend(opts)
        sql = ('SELECT ps.sequence, ppq.quant, pf.fdr, pp.pep '
               'FROM peptide_precur_quanted AS ppq '
               'JOIN peptide_sequences AS ps USING(pep_id) '
               'JOIN peptide_fdr AS pf USING(pep_id) '
               'JOIN peptide_pep AS pp USING(pep_id) '
               )
        checkfeats = ['Peptide sequence', 'MS1 area (highest of all PSMs)',
                      'q-value', 'PEP']
        self.run_checks(options, sql, checkfeats)
        iso_sql = ('SELECT p.sequence, pc.channel_name, piq.quantvalue '
                   'FROM peptide_iso_quanted AS piq '
                   'JOIN peptide_sequences AS p USING(pep_id) '
                   'JOIN pepquant_channels AS pc USING(channel_id)')
        self.check_isobaric(iso_sql)

    def test_genecentric(self):
        self.run_peptable(['--genecentric'])

    def test_proteincentric(self):
        self.run_peptable()


class TestProteintableLookup(TestProtPepTableLookup):
    command = 'proteins'
    infilename = 'full_prottable.txt'
    base_db_fn = 'mzidtsv_db.sqlite'

    def test_proteincentric(self):
        options = ['--setnames', 'S1',
                   '--isobquantcolpattern', 'itraq4plex', '--fdrcolpattern',
                   'q-value', '--pepcolpattern', 'PEP', '--psmnrcolpattern',
                   'quanted', '--ms1quantcolpattern', 'area',
                   '--probcolpattern', 'probability']
        sql = ('SELECT p.protein_acc, ppq.quant, pf.fdr, pp.pep, '
               'ppr.probability '
               'FROM protein_precur_quanted AS ppq '
               'JOIN proteins AS p USING(pacc_id) '
               'JOIN protein_fdr AS pf USING(pacc_id) '
               'JOIN protein_pep AS pp USING(pacc_id) '
               'JOIN protein_probability AS ppr USING(pacc_id)')
        checkfeats = ['Protein accession', 'MS1 precursor area', 'q-value',
                      'PEP', 'Protein error probability']
        self.run_checks(options, sql, checkfeats)
        iso_sql = ('SELECT p.protein_acc, pc.channel_name, piq.quantvalue '
                   'FROM protein_iso_quanted AS piq '
                   'JOIN proteins AS p USING(pacc_id) '
                   'JOIN protquant_channels AS pc USING(channel_id)')
        self.check_isobaric(iso_sql)

    def test_genecentric_no_psmnrs(self):
        options = ['--setnames', 'S1', '--genecentric', 'genes',
                   '--isobquantcolpattern', 'itraq4plex', '--fdrcolpattern',
                   'q-value', '--pepcolpattern', 'PEP', 
                   '--ms1quantcolpattern', 'area', '--protcol', '2',
                   '--probcolpattern', 'probability']
        sql = ('SELECT p.gene_acc, ppq.quant, pf.fdr, pp.pep, '
               'ppr.probability '
               'FROM gene_precur_quanted AS ppq '
               'JOIN genes AS p USING(gene_id) '
               'JOIN gene_fdr AS pf USING(gene_id) '
               'JOIN gene_pep AS pp USING(gene_id) '
               'JOIN gene_probability AS ppr USING(gene_id)')
        checkfeats = ['Gene', 'MS1 precursor area', 'q-value',
                      'PEP', 'Protein error probability']
        self.run_checks(options, sql, checkfeats)
        iso_sql = ('SELECT p.gene_acc, pc.channel_name, piq.quantvalue '
                   'FROM gene_iso_quanted AS piq '
                   'JOIN genes AS p USING(gene_id) '
                   'JOIN genequant_channels AS pc USING(channel_id)')
        self.check_isobaric(iso_sql)

    def test_assoccentric(self):
        options = ['--setnames', 'S1', '--genecentric', 'assoc',
                   '--isobquantcolpattern', 'itraq4plex', '--fdrcolpattern',
                   'q-value', '--pepcolpattern', 'PEP', '--psmnrcolpattern',
                   'quanted', '--ms1quantcolpattern', 'area', '--protcol', '3',
                   '--probcolpattern', 'probability']
        sql = ('SELECT p.assoc_id, ppq.quant, pf.fdr, pp.pep, '
               'ppr.probability '
               'FROM assoc_precur_quanted AS ppq '
               'JOIN associated_ids AS p USING(gene_id) '
               'JOIN assoc_fdr AS pf USING(gene_id) '
               'JOIN assoc_pep AS pp USING(gene_id) '
               'JOIN assoc_probability AS ppr USING(gene_id)')
        checkfeats = ['Associated gene ID', 'MS1 precursor area', 'q-value',
                      'PEP', 'Protein error probability']
        self.run_checks(options, sql, checkfeats)
        iso_sql = ('SELECT p.assoc_id, pc.channel_name, piq.quantvalue '
                   'FROM assoc_iso_quanted AS piq '
                   'JOIN associated_ids AS p USING(gene_id) '
                   'JOIN genequant_channels AS pc USING(channel_id)')
        self.check_isobaric(iso_sql)
